""" Stages that implement utility functions """

import tables_io
from ceci.config import StageParameter as Param

from rail.core.data import DataHandle, Hdf5Handle, PqHandle, TableLike
from rail.core.stage import RailStage


class ColumnMapper(RailStage):
    """Utility stage that remaps the names of columns.

    1. This operates on pandas dataframs in parquet files.

    2. In short, this does:
    `output_data = input_data.rename(columns=self.config.columns, inplace=self.config.inplace)`

    """

    name = "ColumnMapper"
    config_options = RailStage.config_options.copy()
    config_options.update(
        columns=Param(dict, required=True, msg="Map of columns to rename"),
        inplace=Param(bool, default=False, msg="Update file in place"),
    )
    inputs = [("input", PqHandle)]
    outputs = [("output", PqHandle)]

    def run(self) -> None:
        data = self.get_data("input", allow_missing=True)
        out_data = data.rename(columns=self.config.columns, inplace=self.config.inplace)
        if self.config.inplace:  # pragma: no cover
            out_data = data
        self.add_data("output", out_data)

    def __repr__(self) -> str:  # pragma: no cover
        printMsg = "Stage that applies remaps the following column names in a pandas DataFrame:\n"
        printMsg += "f{str(self.config.columns)}"
        return printMsg

    def __call__(self, data: TableLike) -> DataHandle:
        """Return a table with the columns names changed

        Parameters
        ----------
        sample : Table-like
            The data to be renamed

        Returns
        -------
        table : Table-like
            The degraded sample
        """
        self.set_data("input", data)
        self.run()
        self.finalize()
        return self.get_handle("output")


class RowSelector(RailStage):
    """Utility Stage that sub-selects rows from a table by index

    1. This operates on pandas dataframs in parquet files.

    2. In short, this does:
    `output_data = input_data[self.config.start:self.config.stop]`

    """

    name = "RowSelector"
    config_options = RailStage.config_options.copy()
    config_options.update(
        start=Param(int, required=True, msg="Starting row number"),
        stop=Param(int, required=True, msg="Stoppig row number"),
    )
    inputs = [("input", PqHandle)]
    outputs = [("output", PqHandle)]

    def run(self) -> None:
        data = self.get_data("input", allow_missing=True)
        out_data = data.iloc[self.config.start : self.config.stop]
        self.add_data("output", out_data)

    def __repr__(self) -> str:  # pragma: no cover
        printMsg = "Stage that applies remaps the following column names in a pandas DataFrame:\n"
        printMsg += "f{str(self.config.columns)}"
        return printMsg

    def __call__(self, data: TableLike) -> DataHandle:
        """Return a table with the columns names changed

        Parameters
        ----------
        sample : table-like
            The data to be renamed

        Returns
        -------
        table : table-like
            The degraded sample
        """
        self.set_data("input", data)
        self.run()
        self.finalize()
        return self.get_handle("output")


class TableConverter(RailStage):
    """Utility stage that converts tables from one format to anothe

    FIXME, this is hardwired to convert parquet tables to Hdf5Tables.
    It would be nice to have more options here.
    """

    name = "TableConverter"
    config_options = RailStage.config_options.copy()
    config_options.update(
        output_format=Param(str, required=True, msg="Format of output table"),
    )
    inputs = [("input", PqHandle)]
    outputs = [("output", Hdf5Handle)]

    def run(self) -> None:
        data = self.get_data("input", allow_missing=True)
        out_fmt = tables_io.types.TABULAR_FORMAT_NAMES[self.config.output_format]
        out_data = tables_io.convert(data, out_fmt)
        self.add_data("output", out_data)

    def __call__(self, data: TableLike) -> DataHandle:
        """Return a converted table

        Parameters
        ----------
        data : table-like
            The data to be converted

        Returns
        -------
        out_data : table-like
            The converted version of the table
        """
        self.set_data("input", data)
        self.run()
        self.finalize()
        return self.get_handle("output")



"""
" Abstract base class defining an Evaluator

The key feature is that the evaluate method.
"""

from typing import Any, Generator, Iterable

import numpy as np
from ceci.config import StageParameter as Param
from qp.metrics import MetricInputType, MetricOutputType
from qp.metrics.base_metric_classes import BaseMetric

from rail.core.data import QPOrTableHandle
from rail.evaluation.evaluator import Evaluator


class SingleEvaluator(Evaluator):  # pylint: disable=too-many-instance-attributes
    """Evaluate the performance of a photo-Z estimator"""

    name = "SingleEvaluator"
    config_options = Evaluator.config_options.copy()
    config_options.update(
        point_estimates=Param(list, msg="List of point estimates to use", default=[]),
        truth_point_estimates=Param(
            list, msg="List of true point values to use", default=[]
        ),
        hdf5_groupname=Param(
            str, "photometry", required=False, msg="HDF5 Groupname for truth table."
        ),
    )
    inputs = [("input", QPOrTableHandle), ("truth", QPOrTableHandle)]

    metric_base_class = BaseMetric

    def __init__(self, args: Any, **kwargs: Any) -> None:
        """Initialize Evaluator"""
        super().__init__(args, **kwargs)
        self._input_data_type = QPOrTableHandle.PdfOrValue.unknown
        self._truth_data_type = QPOrTableHandle.PdfOrValue.unknown
        self._out_table: dict = {}
        self._summary_table: dict = {}

    def run(self) -> None:  # pylint: disable=too-many-branches
        """Run method

        Evaluate all the metrics and put them into a table

        Notes
        -----
        Get the input data from the data store under this stages 'input' tag
        Get the truth data from the data store under this stages 'truth' tag
        Puts the data into the data store under this stages 'output' tag
        """
        input_data_handle = self.get_handle("input", allow_missing=True)
        truth_data_handle = self.get_handle("truth", allow_missing=True)

        self._input_data_type = input_data_handle.check_pdf_or_point()
        self._truth_data_type = truth_data_handle.check_pdf_or_point()

        Evaluator.run(self)

    def _process_chunk(self, data_tuple: Any, first: bool) -> None:
        start = data_tuple[0]
        end = data_tuple[1]
        input_data = data_tuple[2]
        truth_data = data_tuple[3]

        for metric, this_metric in self._cached_metrics.items():
            if (
                this_metric.metric_input_type == MetricInputType.single_ensemble
            ):  # pragma: no cover
                if not self._input_data_type.has_dist():  # pragma: no cover
                    print(
                        f"skipping {metric} {self._input_data_type} {this_metric.metric_input_type}"
                    )
                    continue
                key_val = f"{metric}"
                self._process_chunk_single_ensemble(this_metric, key_val, input_data)
            elif (
                this_metric.metric_input_type == MetricInputType.dist_to_dist
            ):  # pragma: no cover
                if (
                    not self._input_data_type.has_dist()
                    or not self._truth_data_type.has_dist()
                ):  # pragma: no cover
                    print(
                        f"skipping {metric} {self._input_data_type} {this_metric.metric_input_type}"
                    )
                    continue
                key_val = f"{metric}"
                self._process_chunk_dist_to_dist(
                    this_metric, key_val, input_data, truth_data
                )
            elif this_metric.metric_input_type == MetricInputType.dist_to_point:
                if (
                    not self._input_data_type.has_dist()
                    or not self._truth_data_type.has_point()
                ):  # pragma: no cover
                    print(
                        f"skipping {metric} {self._input_data_type} {this_metric.metric_input_type}"
                    )
                    continue
                for truth_point_estimate_ in self.config.truth_point_estimates:
                    key_val = f"{metric}_{truth_point_estimate_}"
                    self._process_chunk_dist_to_point(
                        this_metric,
                        key_val,
                        input_data,
                        np.squeeze(truth_data[truth_point_estimate_]),
                    )
            elif this_metric.metric_input_type == MetricInputType.point_to_point:
                if (
                    not self._input_data_type.has_point()
                    or not self._truth_data_type.has_point()
                ):  # pragma: no cover
                    print(
                        f"skipping {metric} {self._input_data_type} {this_metric.metric_input_type}"
                    )
                    continue
                for point_estimate_ in self.config.point_estimates:
                    point_data = np.squeeze(input_data.ancil[point_estimate_])
                    for truth_point_estimate_ in self.config.truth_point_estimates:
                        key_val = f"{metric}_{point_estimate_}_{truth_point_estimate_}"
                        self._process_chunk_point_to_point(
                            this_metric,
                            key_val,
                            point_data,
                            np.squeeze(truth_data[truth_point_estimate_]),
                        )
            elif (
                this_metric.metric_input_type == MetricInputType.point_to_dist
            ):  # pragma: no cover
                if (
                    not self._input_data_type.has_point()
                    or not self._truth_data_type.has_dist()
                ):  # pragma: no cover
                    print(
                        f"skipping {metric} {self._input_data_type} {this_metric.metric_input_type}"
                    )
                    continue
                for point_estimate_ in self.config.point_estimates:
                    key_val = f"{metric}_{point_estimate_}"
                    self._process_chunk_point_to_dist(
                        this_metric,
                        key_val,
                        point_data,
                        truth_data[truth_point_estimate_],
                    )

        self._output_table_chunk_data(start, end, self._out_table, first)

    def _process_all(self, data_tuple: Any) -> None:
        input_data = data_tuple[0]
        truth_data = data_tuple[1]

        if self.config.hdf5_groupname:
            truth_data = truth_data[self.config.hdf5_groupname]

        for metric, this_metric in self._cached_metrics.items():
            if (
                this_metric.metric_input_type == MetricInputType.single_ensemble
            ):  # pragma: no cover
                if not self._input_data_type.has_dist():  # pragma: no cover
                    continue
                key_val = f"{metric}"
                self._process_all_single_ensemble(this_metric, key_val, input_data)
            elif (
                this_metric.metric_input_type == MetricInputType.dist_to_dist
            ):  # pragma: no cover
                if (
                    not self._input_data_type.has_dist()
                    or not self._truth_data_type.has_dist()
                ):  # pragma: no cover
                    continue
                key_val = f"{metric}"
                self._process_all_dist_to_dist(
                    this_metric, key_val, input_data, truth_data
                )
            elif this_metric.metric_input_type == MetricInputType.dist_to_point:
                if (
                    not self._input_data_type.has_dist()
                    or not self._truth_data_type.has_point()
                ):  # pragma: no cover
                    continue
                for truth_point_estimate_ in self.config.truth_point_estimates:
                    key_val = f"{metric}_{truth_point_estimate_}"
                    self._process_all_dist_to_point(
                        this_metric,
                        key_val,
                        input_data,
                        truth_data[truth_point_estimate_],
                    )
            elif this_metric.metric_input_type == MetricInputType.point_to_point:
                if (
                    not self._input_data_type.has_point()
                    or not self._truth_data_type.has_point()
                ):  # pragma: no cover
                    continue
                for point_estimate_ in self.config.point_estimates:
                    point_data = input_data.ancil[point_estimate_]
                    for truth_point_estimate_ in self.config.truth_point_estimates:
                        key_val = f"{metric}_{point_estimate_}_{truth_point_estimate_}"
                        self._process_all_point_to_point(
                            this_metric,
                            key_val,
                            point_data,
                            truth_data[truth_point_estimate_],
                        )
            elif (
                this_metric.metric_input_type == MetricInputType.point_to_dist
            ):  # pragma: no cover
                if (
                    not self._input_data_type.has_point()
                    or not self._truth_data_type.has_dist()
                ):  # pragma: no cover
                    continue
                for point_estimate_ in self.config.point_estimates:
                    key_val = f"{metric}_{point_estimate_}"
                    self._process_all_point_to_dist(
                        this_metric,
                        key_val,
                        point_data,
                        truth_data[truth_point_estimate_],
                    )

        out_table_to_write = {
            key: np.array(val).astype(float) for key, val in self._out_table.items()
        }
        self._output_handle = self.add_handle("output", data=out_table_to_write)
        self._summary_handle = self.add_handle("summary", data=self._summary_table)
        self._single_distribution_summary_handle = self.add_handle(
            "single_distribution_summary", data=self._single_distribution_summary_data
        )

    def _process_chunk_single_ensemble(
        self, this_metric: BaseMetric, key: str, input_data: Any
    ) -> None:  # pragma: no cover
        if this_metric.metric_output_type == MetricOutputType.single_value:
            if not hasattr(this_metric, "accumulate"):
                print(
                    f"{this_metric.metric_name} with output type "
                    "single_value does not support parallel processing"
                )
                return

            centroids = this_metric.accumulate(input_data)

            if self.comm:
                self._cached_data[key] = centroids
            else:
                if key in self._cached_data:
                    self._cached_data[key].append(centroids)
                else:
                    self._cached_data[key] = [centroids]

        elif this_metric.metric_output_type == MetricOutputType.single_distribution:
            if not hasattr(this_metric, "accumulate"):  # pragma: no cover
                print(
                    f"{this_metric.metric_name} with output type "
                    "MetricOutputType.single_value does not support parallel processing yet"
                )
                return

            accumulated_data = this_metric.accumulate(input_data)
            if self.comm:
                self._cached_data[key] = accumulated_data
            else:
                if key in self._cached_data:
                    self._cached_data[key].append(accumulated_data)
                else:
                    self._cached_data[key] = [accumulated_data]
        else:
            self._out_table[key] = this_metric.evaluate(input_data)

    def _process_chunk_dist_to_dist(
        self, this_metric: BaseMetric, key: str, input_data: Any, truth_data: Any
    ) -> None:  # pragma: no cover
        if this_metric.metric_output_type == MetricOutputType.single_value:
            if not hasattr(this_metric, "accumulate"):
                print(
                    f"{this_metric.metric_name} with output type "
                    "single_value does not support parallel processing yet"
                )
                return

            centroids = this_metric.accumulate(input_data, truth_data)
            if self.comm:
                self._cached_data[key] = centroids
            else:
                if key in self._cached_data:
                    self._cached_data[key].append(centroids)
                else:
                    self._cached_data[key] = [centroids]

        elif this_metric.metric_output_type == MetricOutputType.single_distribution:
            if not hasattr(this_metric, "accumulate"):  # pragma: no cover
                print(
                    f"{this_metric.metric_name} with output type "
                    "MetricOutputType.single_value does not support parallel processing yet"
                )
                return

            accumulated_data = this_metric.accumulate(input_data, truth_data)
            if self.comm:
                self._cached_data[key] = accumulated_data
            else:
                if key in self._cached_data:
                    self._cached_data[key].append(accumulated_data)
                else:
                    self._cached_data[key] = [accumulated_data]

        else:
            self._out_table[key] = this_metric.evaluate(input_data, truth_data)

    def _process_chunk_dist_to_point(
        self, this_metric: BaseMetric, key: str, input_data: Any, truth_data: Any
    ) -> None:
        if this_metric.metric_output_type == MetricOutputType.single_value:
            if not hasattr(this_metric, "accumulate"):  # pragma: no cover
                print(
                    f"{this_metric.metric_name} with output type "
                    "single_value does not support parallel processing yet"
                )
                return

            centroids = this_metric.accumulate(input_data, truth_data)
            if self.comm:  # pragma: no cover
                self._cached_data[key] = centroids
            else:
                if key in self._cached_data:
                    self._cached_data[key].append(centroids)
                else:
                    self._cached_data[key] = [centroids]

        elif (
            this_metric.metric_output_type == MetricOutputType.single_distribution
        ):  # pragma: no cover
            if not hasattr(this_metric, "accumulate"):
                print(
                    f"{this_metric.metric_name} with output type "
                    "single_value does not support parallel processing yet"
                )
                return
            accumulated_data = this_metric.accumulate(input_data, truth_data)
            if self.comm:
                self._cached_data[key] = accumulated_data
            else:
                if key in self._cached_data:
                    self._cached_data[key].append(accumulated_data)
                else:
                    self._cached_data[key] = [accumulated_data]

        else:  # pragma: no cover
            self._out_table[key] = this_metric.evaluate(input_data, truth_data)

    def _process_chunk_point_to_dist(
        self, this_metric: BaseMetric, key: str, input_data: Any, truth_data: Any
    ) -> None:  # pragma: no cover
        if this_metric.metric_output_type == MetricOutputType.single_value:
            if not hasattr(this_metric, "accumulate"):  # pragma: no cover
                print(
                    f"{this_metric.metric_name} with output type "
                    "single_value does not support parallel processing yet"
                )
                return

            centroids = this_metric.accumulate(input_data, truth_data)
            if self.comm:  # pragma: no cover
                self._cached_data[key] = centroids
            else:
                if key in self._cached_data:
                    self._cached_data[key].append(centroids)
                else:
                    self._cached_data[key] = [centroids]

        elif this_metric.metric_output_type == MetricOutputType.single_distribution:
            if not hasattr(this_metric, "accumulate"):  # pragma: no cover
                print(
                    f"{this_metric.metric_name} with output type "
                    "single_value does not support parallel processing yet"
                )
                return
            accumulated_data = this_metric.accumulate(input_data, truth_data)
            if self.comm:
                self._cached_data[key] = accumulated_data
            else:
                if key in self._cached_data:
                    self._cached_data[key].append(accumulated_data)
                else:
                    self._cached_data[key] = [accumulated_data]
        else:
            self._out_table[key] = this_metric.evaluate(input_data, truth_data)

    def _process_chunk_point_to_point(
        self, this_metric: BaseMetric, key: str, input_data: Any, truth_data: Any
    ) -> None:
        if this_metric.metric_output_type == MetricOutputType.single_value:
            if not hasattr(this_metric, "accumulate"):  # pragma: no cover
                print(
                    f"{this_metric.metric_name} with output type "
                    "single_value does not support parallel processing yet"
                )
                return

            centroids = this_metric.accumulate(input_data, truth_data)
            if self.comm:  # pragma: no cover
                self._cached_data[key] = centroids
            else:
                if key in self._cached_data:
                    self._cached_data[key].append(centroids)
                else:
                    self._cached_data[key] = [centroids]

        elif (
            this_metric.metric_output_type == MetricOutputType.single_distribution
        ):  # pragma: no cover
            if not hasattr(this_metric, "accumulate"):
                print(
                    f"{this_metric.metric_name} with output type "
                    "single_value does not support parallel processing yet"
                )
                return
            accumulated_data = this_metric.accumulate(input_data, truth_data)
            if self.comm:
                self._cached_data[key] = accumulated_data
            else:
                if key in self._cached_data:
                    self._cached_data[key].append(accumulated_data)
                else:
                    self._cached_data[key] = [accumulated_data]
        else:
            self._out_table[key] = this_metric.evaluate(input_data, truth_data)

    def _process_all_single_ensemble(
        self, this_metric: BaseMetric, key: str, input_data: Any
    ) -> None:  # pragma: no cover
        if this_metric.metric_output_type == MetricOutputType.single_value:
            self._summary_table[key] = np.array([this_metric.evaluate(input_data)])

        elif this_metric.metric_output_type == MetricOutputType.single_distribution:
            self._single_distribution_summary_data[key] = this_metric.evaluate(
                input_data
            )
        else:
            self._out_table[key] = this_metric.evaluate(input_data)

    def _process_all_dist_to_dist(
        self, this_metric: BaseMetric, key: str, input_data: Any, truth_data: Any
    ) -> None:  # pragma: no cover
        if this_metric.metric_output_type == MetricOutputType.single_value:
            self._summary_table[key] = np.array(
                [this_metric.evaluate(input_data, truth_data)]
            )

        elif this_metric.metric_output_type == MetricOutputType.single_distribution:
            self._single_distribution_summary_data[key] = this_metric.evaluate(
                input_data, truth_data
            )
        else:
            self._out_table[key] = this_metric.evaluate(input_data, truth_data)

    def _process_all_dist_to_point(
        self, this_metric: BaseMetric, key: str, input_data: Any, truth_data: Any
    ) -> None:
        if this_metric.metric_output_type == MetricOutputType.single_value:
            self._summary_table[key] = np.array(
                [this_metric.evaluate(input_data, truth_data)]
            )

        elif (
            this_metric.metric_output_type == MetricOutputType.single_distribution
        ):  # pragma: no cover
            self._single_distribution_summary_data[key] = this_metric.evaluate(
                input_data, truth_data
            )
        else:  # pragma: no cover
            self._out_table[key] = this_metric.evaluate(input_data, truth_data)

    def _process_all_point_to_dist(
        self, this_metric: BaseMetric, key: str, input_data: Any, truth_data: Any
    ) -> None:  # pragma: no cover
        if this_metric.metric_output_type == MetricOutputType.single_value:
            self._summary_table[key] = np.array(
                [this_metric.evaluate(input_data, truth_data)]
            )

        elif this_metric.metric_output_type == MetricOutputType.single_distribution:
            self._single_distribution_summary_data[key] = this_metric.evaluate(
                input_data, truth_data
            )
        else:
            self._out_table[key] = this_metric.evaluate(input_data, truth_data)

    def _process_all_point_to_point(
        self, this_metric: BaseMetric, key: str, input_data: Any, truth_data: Any
    ) -> None:
        if this_metric.metric_output_type == MetricOutputType.single_value:
            self._summary_table[key] = np.array(
                [this_metric.evaluate(input_data, truth_data)]
            )

        elif (
            this_metric.metric_output_type == MetricOutputType.single_distribution
        ):  # pragma: no cover
            self._single_distribution_summary_data[key] = this_metric.evaluate(
                input_data, truth_data
            )
        else:
            self._out_table[key] = this_metric.evaluate(input_data, truth_data)

    def _setup_iterator(self, itrs: list[Iterable] | None = None) -> Generator:
        if itrs is None:
            tags = ["input", "truth"]
            itrs = [
                self.input_iterator(tag, groupname=self.config.hdf5_groupname)
                for tag in tags
            ]
        return Evaluator._setup_iterator(self, itrs)



import tempfile

import rail
import rail.stages
from rail.core import RailEnv
from rail.core.data import DataHandle, TableHandle


def test_print_rail_packages() -> None:
    RailEnv.print_rail_packages()


def test_print_rail_namespaces() -> None:
    RailEnv.print_rail_namespaces()


def test_print_rail_modules() -> None:
    RailEnv.print_rail_modules()


def test_print_rail_namespace_tree() -> None:
    RailEnv.print_rail_namespace_tree()


def test_import_and_attach_all() -> None:
    rail.stages.import_and_attach_all()
    RailEnv.print_rail_stage_dict()


def test_api_rst() -> None:
    with tempfile.TemporaryDirectory() as tmpdirname:
        RailEnv.do_api_rst(tmpdirname)
        RailEnv.do_stage_type_api_rst(tmpdirname)


def test_data_handle_dict() -> None:
    DataHandle.print_sub_classes()
    assert DataHandle.get_sub_class("TableHandle") == TableHandle



"""
Abstract base class defining an Evaluator

The key feature is that the evaluate method.
"""

from typing import Any, Generator, Iterable

import numpy as np
import qp
from ceci.config import StageParameter as Param
from ceci.stage import PipelineStage
from qp.metrics.base_metric_classes import BaseMetric, MetricOutputType
from qp.metrics.pit import PIT

from rail.core.common_params import SHARED_PARAMS
from rail.core.data import DataHandle, Hdf5Handle, QPDictHandle, QPHandle
from rail.core.stage import RailStage
from rail.evaluation.metrics.cdeloss import CDELoss
from rail.evaluation.metrics.pointestimates import (PointBias,
                                                    PointOutlierRate,
                                                    PointSigmaIQR,
                                                    PointSigmaMAD)


def _all_subclasses(a_class: type[BaseMetric]) -> set[type[BaseMetric]]:
    return set(a_class.__subclasses__()).union(
        [s for c in a_class.__subclasses__() for s in _all_subclasses(c)]
    )


def _build_metric_dict(a_class: type[BaseMetric]) -> dict[str, type[BaseMetric]]:
    the_dict: dict[str, type[BaseMetric]] = {}
    for subcls in _all_subclasses(a_class):
        if subcls.metric_name is None:
            continue
        the_dict[subcls.metric_name] = subcls
    return the_dict


class Evaluator(RailStage):  # pylint: disable=too-many-instance-attributes
    """Evaluate the performance of a photo-z estimator against reference point estimate"""

    name = "Evaluator"
    config_options = RailStage.config_options.copy()
    config_options.update(
        metrics=Param(
            list, [], required=False, msg="The metrics you want to evaluate."
        ),
        exclude_metrics=Param(
            list, [], msg="List of metrics to exclude", required=False
        ),
        metric_config=Param(
            dict, msg="configuration of individual_metrics", default={}
        ),
        chunk_size=Param(
            int,
            10000,
            required=False,
            msg="The default number of PDFs to evaluate per loop.",
        ),
        _random_state=Param(
            float,
            default=None,
            required=False,
            msg="Random seed value to use for reproducible results.",
        ),
        force_exact=Param(
            bool,
            default=False,
            required=False,
            msg="Force the exact calculation.  This will not allow parallelization",
        ),
    )

    inputs: list[tuple[str, type[DataHandle]]] = []
    outputs = [
        ("output", Hdf5Handle),
        ("summary", Hdf5Handle),
        ("single_distribution_summary", QPDictHandle),
    ]

    metric_base_class: type[BaseMetric] | None = None

    def __init__(self, args: Any, **kwargs: Any) -> None:
        super().__init__(args, **kwargs)
        self._output_handle: DataHandle | None = None
        self._summary_handle: DataHandle | None = None
        self._single_distribution_summary_handle: DataHandle | None = None
        assert self.metric_base_class is not None
        self._metric_dict = _build_metric_dict(self.metric_base_class)
        self._cached_data: dict = {}
        self._cached_metrics: dict = {}
        self._single_distribution_summary_data: dict = {}
        self._metric_config_dict: dict = {}

    def evaluate(self, data: qp.Ensemble, truth: Any) -> dict[str, DataHandle]:
        """Evaluate the performance of an estimator

        This will attach the input data and truth to this `Evaluator`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the data that it creates to this Estimator
        by using `self.add_data('output', output_data)`.

        Parameters
        ----------
        data
            The sample to evaluate

        truth
            Table with the truth information

        Returns
        -------
        dict[str, DataHandle]
            The evaluation metrics
        """

        self.set_data("input", data)
        self.set_data("truth", truth)
        self.run()
        self.finalize()
        return {
            "output": self.get_handle("output"),
            "summary": self.get_handle("summary"),
            "single_distribution_summary": self.get_handle(
                "single_distribution_summary"
            ),
        }

    def run(self) -> None:
        self._build_config_dict()

        print(f"Requested metrics: {list(self._metric_config_dict.keys())}")

        if self.config.force_exact:
            self.run_single_node()
            return

        itr = self._setup_iterator()

        first = True
        for data_tuple in itr:
            chunk_start, chunk_end = data_tuple[0], data_tuple[1]

            print(
                f"Processing {self.rank} running evaluator on chunk {chunk_start} - {chunk_end}."
            )
            self._process_chunk(data_tuple, first)
            first = False

    def run_single_node(self) -> None:
        data_tuple = self._get_all_data()
        self._process_all(data_tuple)

    def finalize(self) -> None:
        if not self.config.force_exact:
            # Finalize all the metrics that produce a single value summary
            summary_data = {}
            single_distribution_summary_data = {}
            for metric, cached_metric in self._cached_metrics.items():
                if cached_metric.metric_output_type not in [
                    MetricOutputType.single_value,
                    MetricOutputType.single_distribution,
                ]:
                    continue
                matching_keys = []
                for key_ in self._cached_data:
                    if key_.find(metric) == 0:
                        matching_keys.append(key_)
                if not matching_keys:  # pragma: no cover
                    print(
                        f"Skipping {metric} which did not cache data {list(self._cached_data.keys())}"
                    )
                    continue
                for key_ in matching_keys:
                    if self.comm:  # pragma: no cover
                        self._cached_data[key_] = self.comm.gather(
                            self._cached_data[key_]
                        )

                    if (
                        cached_metric.metric_output_type
                        == MetricOutputType.single_value
                    ):
                        summary_data[key_] = np.array(
                            [cached_metric.finalize(self._cached_data[key_])]
                        )

                    elif (
                        cached_metric.metric_output_type
                        == MetricOutputType.single_distribution
                    ):
                        # we expect `cached_metric.finalize` to return a qp.Ensemble
                        single_distribution_summary_data[key_] = cached_metric.finalize(
                            self._cached_data[key_]
                        )

            self._summary_handle = self.add_handle("summary", data=summary_data)
            self._single_distribution_summary_handle = self.add_handle(
                "single_distribution_summary", data=single_distribution_summary_data
            )

        PipelineStage.finalize(self)

    def _setup_iterator(self, itrs: list[Iterable] | None = None) -> Generator:
        """Setup the iterator that runs in parallel over the handles"""

        if itrs is None:
            handle_list = [
                input_[0] for input_ in self.inputs
            ]  # pylint: disable=no-member
            itrs = [self.input_iterator(tag) for tag in handle_list]

        for it in zip(*itrs):
            data = []
            first = True
            for s, e, d in it:
                if first:
                    data.append(s)
                    data.append(e)
                    data.append(d)
                    first = False
                else:
                    data.append(d)
            yield data

    def _get_all_data(self) -> Any:
        """Stuff the data from all the handles into a tuple"""
        handles = [input_[0] for input_ in self.inputs]  # pylint: disable=no-member
        all_data = [self.get_data(handle_) for handle_ in handles]
        return all_data

    def _process_chunk(self, data_tuple: Any, first: bool) -> None:  # pragma: no cover
        raise NotImplementedError("Evaluator._process_chunk()")

    def _process_all_chunk_metrics(
        self, estimate_data: Any, reference_data: Any, start: int, end: int, first: bool
    ) -> None:
        """This function takes the properly formatted data and loops over all the
        requested metrics. The metric outputs are either cached for later finalization
        or written to output files.

        Parameters
        ----------
        estimate_data
            The estimated values (of the appropriate type, float or pdf) to be used
            by the requested metrics.

        reference_data
            The reference or known values (of the appropriate type, float or pdf)
            to be used by the requested metrics.

        start
            The start index of the data chunk, used to write metric results to
            the correct location of the output file.

        end
            The end index of the data chunk, used to write metric results to the
            correct location of the output file.

        first
            Used internally to determine if the output file should be initialized.

        Raises
        ------
        ValueError
            Raises an error if an unknown metric is requested.
        """
        out_table = {}
        for metric, this_metric in self._cached_metrics.items():
            if metric not in self._metric_dict:  # pragma: no cover
                raise ValueError(
                    f"Unsupported metric requested: '{metric}'. "
                    f"Available metrics are: {sorted(self._metric_dict.keys())}"
                )

            if this_metric.metric_output_type in [
                MetricOutputType.single_value,
                MetricOutputType.single_distribution,
            ]:
                if not hasattr(this_metric, "accumulate"):  # pragma: no cover
                    print(
                        f"The metric `{metric}` does not support parallel processing yet"
                    )
                    continue

                centroids = this_metric.accumulate(estimate_data, reference_data)
                if self.comm:  # pragma: no cover
                    self._cached_data[metric] = centroids
                else:
                    if metric in self._cached_data:
                        self._cached_data[metric].append(centroids)
                    else:
                        self._cached_data[metric] = [centroids]

            else:
                out_table[metric] = this_metric.evaluate(
                    estimate_data,
                    reference_data,
                )

        self._output_table_chunk_data(start, end, out_table, first)

    def _output_table_chunk_data(
        self, start: int, end: int, out_table: Any, first: bool
    ) -> None:
        out_table_to_write = {
            key: np.array(val).astype(float) for key, val in out_table.items()
        }

        if first:
            self._output_handle = self.add_handle("output", data=out_table_to_write)
            assert isinstance(self._output_handle, DataHandle)
            self._output_handle.initialize_write(
                self._input_length, communicator=self.comm
            )
        assert self._output_handle is not None
        self._output_handle.set_data(out_table_to_write, partial=True)
        self._output_handle.write_chunk(start, end)

    def _process_all(self, data_tuple: Any) -> None:  # pragma: no cover
        raise NotImplementedError("Evaluator._process_all()")

    def _process_all_metrics(self, estimate_data: Any, reference_data: Any) -> None:
        """This function writes out metric values when operating in non-parallel mode.

        Parameters
        ----------
        estimate_data
            The estimated values (of the appropriate type, float or pdf) to be used
            by the requested metrics.

        reference_data
            The reference or known values (of the appropriate type, float or pdf)
            to be used by the requested metrics.

        Raises
        ------
        ValueError
            Raises an error if an unknown metric is requested.
        """

        out_table = {}
        summary_table = {}
        single_distribution_summary = {}

        for metric, this_metric in self._cached_metrics.items():
            if metric not in self._metric_dict:  # pragma: no cover
                raise ValueError(
                    f"Unsupported metric requested: '{metric}'. "
                    f"Available metrics are: {sorted(self._metric_dict.keys())}"
                )

            metric_result = this_metric.evaluate(estimate_data, reference_data)

            if this_metric.metric_output_type == MetricOutputType.single_value:
                summary_table[metric] = np.array([metric_result])
            elif this_metric.metric_output_type == MetricOutputType.single_distribution:
                single_distribution_summary[this_metric.metric_name] = metric_result
            else:
                out_table[metric] = metric_result

        out_table_to_write = {
            key: np.array(val).astype(float) for key, val in out_table.items()
        }
        self._output_handle = self.add_handle("output", data=out_table_to_write)
        self._summary_handle = self.add_handle("summary", data=summary_table)
        self._single_distribution_summary_handle = self.add_handle(
            "single_distribution_summary", data=single_distribution_summary
        )

    def _build_config_dict(self) -> None:
        """Build the configuration dict for each of the metrics"""
        self._metric_config_dict = {}

        if "all" in self.config.metrics:  # pragma: no cover
            metric_list = list(self._metric_dict.keys())
            for exclude_ in self.config.exclude_metrics:
                metric_list.remove(exclude_)
        else:
            metric_list = self.config.metrics

        for metric_name_ in metric_list:
            if metric_name_ in self.config.exclude_metrics:  # pragma: no cover
                continue
            if metric_name_ not in self._metric_dict:
                print(
                    f"Unsupported metric requested: '{metric_name_}'.  "
                    f"Available metrics are: {sorted(self._metric_dict.keys())}"
                )
                continue
            sub_dict = {}
            if "limits" in self.config:
                sub_dict["limits"] = self.config.limits
            sub_dict.update(self.config.metric_config.get("general", {}))
            sub_dict.update(self.config.metric_config.get(metric_name_, {}))
            self._metric_config_dict[metric_name_] = sub_dict
            this_metric_class = self._metric_dict[metric_name_]
            try:
                this_metric = this_metric_class(**sub_dict)
            except (TypeError, KeyError):  # pragma: no cover
                sub_dict.pop("limits")
                this_metric = this_metric_class(**sub_dict)
            self._cached_metrics[metric_name_] = this_metric


class OldEvaluator(RailStage):
    """Evaluate the performance of a photo-Z estimator"""

    name = "OldEvaluator"
    config_options = RailStage.config_options.copy()
    config_options.update(
        zmin=Param(float, 0.0, msg="min z for grid"),
        zmax=Param(float, 3.0, msg="max z for grid"),
        nzbins=Param(int, 301, msg="# of bins in zgrid"),
        pit_metrics=Param(str, "all", msg="PIT-based metrics to include"),
        point_metrics=Param(str, "all", msg="Point-estimate metrics to include"),
        hdf5_groupname=Param(
            str, "", msg="Name of group in hdf5 where redshift data is located"
        ),
        do_cde=Param(bool, True, msg="Evaluate CDE Metric"),
        redshift_col=SHARED_PARAMS,
    )
    inputs = [("input", QPHandle), ("truth", Hdf5Handle)]
    outputs = [("output", Hdf5Handle)]

    def evaluate(self, data: qp.Ensemble, truth: Any) -> DataHandle:
        """Evaluate the performance of an estimator

        This will attach the input data and truth to this `Evaluator`
        (for introspection and provenance tracking).
        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.
        The run() method will need to register the data that it creates to this Estimator
        by using `self.add_data('output', output_data)`.

        Parameters
        ----------
        data
            The sample to evaluate

        truth
            Table with the truth information

        Returns
        -------
        DataHandle
            The evaluation metrics
        """

        self.set_data("input", data)
        self.set_data("truth", truth)
        self.run()
        self.finalize()
        return self.get_handle("output")

    def run(self) -> None:
        """Run method
        Evaluate all the metrics and put them into a table
        Notes
        -----
        Get the input data from the data store under this stages 'input' tag
        Get the truth data from the data store under this stages 'truth' tag
        Puts the data into the data store under this stages 'output' tag
        """

        pz_data = self.get_data("input")
        if self.config.hdf5_groupname:  # pragma: no cover
            specz_data = self.get_data("truth")[self.config.hdf5_groupname]
        else:
            specz_data = self.get_data("truth")
        z_true = specz_data[self.config["redshift_col"]]

        zgrid = np.linspace(self.config.zmin, self.config.zmax, self.config.nzbins + 1)

        # Create an instance of the PIT class
        pitobj = PIT(pz_data, z_true)

        # Build reference dictionary of the PIT meta-metrics from this PIT instance
        PIT_METRICS = dict(
            AD=getattr(pitobj, "evaluate_PIT_anderson_ksamp"),
            CvM=getattr(pitobj, "evaluate_PIT_CvM"),
            KS=getattr(pitobj, "evaluate_PIT_KS"),
            OutRate=getattr(pitobj, "evaluate_PIT_outlier_rate"),
        )

        # Parse the input configuration to determine which meta-metrics should be calculated
        if self.config.pit_metrics == "all":
            pit_metrics = list(PIT_METRICS.keys())
        else:  # pragma: no cover
            pit_metrics = self.config.pit_metrics.split()

        # Evaluate each of the requested meta-metrics, and store the result in `out_table`
        out_table = {}
        for pit_metric in pit_metrics:
            value = PIT_METRICS[pit_metric]()
            # The result objects of some meta-metrics are bespoke scipy objects with inconsistent fields.
            # Here we do our best to store the relevant fields in `out_table`.
            if isinstance(value, list):  # pragma: no cover
                out_table[f"PIT_{pit_metric}"] = value
            elif pit_metric == "OutRate":
                out_table[f"PIT_{pit_metric}_stat"] = [value]
            else:
                out_table[f"PIT_{pit_metric}_stat"] = [
                    getattr(value, "statistic", None)
                ]
                if getattr(value, "pvalue", None) is not None:
                    out_table[f"PIT_{pit_metric}_pval"] = [
                        getattr(value, "p_value", None)
                    ]
                if getattr(value, "significance_level", None) is not None:
                    out_table[f"PIT_{pit_metric}_significance_level"] = [
                        getattr(value, "significance_level", None)
                    ]

        POINT_METRICS = dict(
            SimgaIQR=PointSigmaIQR,
            Bias=PointBias,
            OutlierRate=PointOutlierRate,
            SigmaMAD=PointSigmaMAD,
        )
        if self.config.point_metrics == "all":
            point_metrics = list(POINT_METRICS.keys())
        else:  # pragma: no cover
            point_metrics = self.config.point_metrics.split()

        z_mode = None
        for point_metric in point_metrics:
            if z_mode is None:
                z_mode = np.squeeze(pz_data.mode(grid=zgrid))
            value = POINT_METRICS[point_metric](z_mode, z_true).evaluate()
            out_table[f"POINT_{point_metric}"] = [value]

        if self.config.do_cde:
            value = CDELoss(pz_data, zgrid, z_true).evaluate()
            out_table["CDE_stat"] = [value.statistic]
            # out_table["CDE_pval"] = [value.p_value]

        # Converting any possible None to NaN to write it
        out_table_to_write = {
            key: np.array(val).astype(float) for key, val in out_table.items()
        }
        self.add_data("output", out_table_to_write)



# pylint: disable=invalid-name
from __future__ import annotations

import enum


class DistributionType(enum.Enum):
    """ Keep track of a particular distribution represents """
    
    ad_hoc = 0
    posterior = 1
    likelihood = 2







import os

import pytest

from rail.interfaces import ToolFactory
from rail.utils.path_utils import find_rail_file


def test_tool_factory() -> None:
    stage = ToolFactory.build_tool_stage(
        "pq_to_hdf",
        "TableConverter",
        "rail.tools.table_tools",
        "dummy.in",
        output_format="numpyDict",
    )

    input_file = find_rail_file("examples_data/testdata/rubin_dm_dc2_example.pq")

    _out_handle = ToolFactory.run_tool_stage(
        stage,
        input_file,
    )

    check_stage = ToolFactory.get_tool_stage("pq_to_hdf")
    assert check_stage == stage

    with pytest.raises(KeyError):
        ToolFactory.get_tool_stage("nope")

    ToolFactory.reset()
    assert not ToolFactory._stage_dict

    try:
        os.unlink("inprogress_output_pq_to_hdf.hdf5")
    except FileNotFoundError:
        pass
    try:
        os.unlink("output_pq_to_hdf.hdf5")
    except FileNotFoundError:
        pass



"""
A summarizer that simple makes a histogram of a point estimate
"""

from typing import Any, Generator

import numpy as np
import qp
from ceci.config import StageParameter as Param

from rail.core.data import DataHandle, QPHandle, TableHandle, TableLike
from rail.estimation.informer import PzInformer
from rail.estimation.estimator import PzEstimator


class GaussianPzInformer(PzInformer):
    """Placeholder Informer"""

    name = "GaussianPzInformer"
    config_options = PzInformer.config_options.copy()

    def _finalize_run(self) -> None:
        self.model = np.array([None])
        PzInformer._finalize_run(self)


class GaussianPzEstimator(PzEstimator):
    """Estimator which converts to Gaussian reps"""

    name = "GaussianPzEstimator"
    config_options = PzEstimator.config_options.copy()

    def _process_chunk(
        self, start: int, end: int, data: qp.Ensemble, first: bool
    ) -> None:

        mean = np.squeeze(data.mean())
        std = np.squeeze(data.std())
        std = np.where(np.isfinite(std), std, 0.01)
        
        qp_dstn = qp.Ensemble(qp.stats.norm, data=dict(loc=mean, scale=std))
        self._do_chunk_output(qp_dstn, start, end, first)

        



"""Abstract base class defining a degrader.

The key feature is that the __call__ method takes a pandas DataFrame and a seed,
and returns a pandas DataFrame, and wraps the run method.
"""

from ceci.config import StageParameter as Param

from rail.core.data import DataHandle, PqHandle, TableLike
from rail.core.stage import RailStage


class Degrader(RailStage):  # pragma: no cover
    """Base class Degraders, which apply various degradations to synthetic
    photometric data.

    Degraders take "input" data in the form of pandas dataframes in Parquet
    files and provide as "output" another pandas dataframes written to Parquet
    files.
    """

    name = "Degrader"
    config_options = RailStage.config_options.copy()
    config_options.update(
        seed=Param(int, default=12345, msg="Random number seed"),
    )
    inputs = [("input", PqHandle)]
    outputs = [("output", PqHandle)]

    def __call__(self, sample: TableLike, seed: int | None = None) -> DataHandle:
        """The main interface method for ``Degrader``.

        Applies degradation.

        This will attach the sample to this `Degrader` (for introspection and
        provenance tracking).

        Then it will call the run() and finalize() methods, which need to be
        implemented by the sub-classes.

        The run() method will need to register the data that it creates to this
        Estimator by using ``self.add_data('output', output_data)``.

        Finally, this will return a PqHandle providing access to that output
        data.

        Parameters
        ----------
        sample
            The sample to be degraded

        seed
            An integer to set the numpy random seed

        Returns
        -------
        DataHandle
            A handle giving access to a table with degraded sample
        """
        if seed is not None:
            self.config.seed = seed
        self.set_data("input", sample)
        self.run()
        self.finalize()
        return self.get_handle("output")



"""Tuples that capture standard statistical quantities """

from collections import namedtuple

# These generic mathematical metrics will be moved to qp at some point.
stat_and_pval = namedtuple("stat_and_pval", ["statistic", "p_value"])
stat_crit_sig = namedtuple(
    "stat_crit_sig", ["statistic", "critical_values", "significance_level"]
)



"""Utility functions to test alogrithms"""

import os
from typing import Any

import ceci

from rail.cli.rail.scripts import build_pipeline
from rail.core.data import TableHandle
from rail.core.stage import RailStage
from rail.estimation.estimator import CatEstimator
from rail.estimation.informer import CatInformer
from rail.utils.path_utils import RAILDIR

traindata = os.path.join(RAILDIR, "rail/examples_data/testdata/training_100gal.hdf5")
validdata = os.path.join(RAILDIR, "rail/examples_data/testdata/validation_10gal.hdf5")
DS = RailStage.data_store
DS.__class__.allow_overwrite = True


def one_algo(
    key: str,
    single_trainer: type[CatInformer],
    single_estimator: type[CatEstimator],
    train_kwargs: dict,
    estim_kwargs: dict,
    is_classifier: bool = False,
) -> Any:
    """
    A basic test of running an estimator subclass.
    Run inform, write temporary trained model to
    'tempmodelfile.tmp', run photo-z algorithm.
    Then, load tempmodelfile.tmp and re-run, return
    both datasets.
    """
    DS.clear()
    training_data = DS.read_file("training_data", TableHandle, traindata)
    validation_data = DS.read_file("validation_data", TableHandle, validdata)

    if single_trainer is not None:
        train_pz = single_trainer.make_stage(**train_kwargs)
        train_pz.inform(training_data)

    pz = single_estimator.make_stage(name=key, **estim_kwargs)
    if not is_classifier:
        estim = pz.estimate(validation_data)
    elif is_classifier:  # pragma: no cover
        estim = pz.classify(validation_data)
    pz_2 = None
    estim_2 = estim
    pz_3 = None
    estim_3 = estim

    copy_estim_kwargs = estim_kwargs.copy()
    model_file = copy_estim_kwargs.pop("model", "None")

    if model_file != "None":
        copy_estim_kwargs["model"] = model_file
        pz_2 = single_estimator.make_stage(name=f"{pz.name}_copy", **copy_estim_kwargs)
        if not is_classifier:
            estim_2 = pz_2.estimate(validation_data)
        elif is_classifier:  # pragma: no cover
            estim_2 = pz_2.classify(validation_data)

    if single_trainer is not None and "model" in single_trainer.output_tags():
        copy3_estim_kwargs = estim_kwargs.copy()
        copy3_estim_kwargs["model"] = train_pz.get_handle("model")
        pz_3 = single_estimator.make_stage(
            name=f"{pz.name}_copy3", **copy3_estim_kwargs
        )
        if not is_classifier:
            estim_3 = pz_3.estimate(validation_data)
        elif is_classifier:  # pragma: no cover
            estim_3 = pz_3.classify(validation_data)

    os.remove(pz.get_output(pz.get_aliased_tag("output"), final_name=True))
    if pz_2 is not None:
        os.remove(pz_2.get_output(pz_2.get_aliased_tag("output"), final_name=True))

    if pz_3 is not None:
        os.remove(pz_3.get_output(pz_3.get_aliased_tag("output"), final_name=True))
    model_file = estim_kwargs.get("model", "None")
    if model_file != "None":
        try:
            os.remove(model_file)
        except FileNotFoundError:  # pragma: no cover
            pass
    return estim.data, estim_2.data, estim_3.data


def check_stage_params(stage_class: type[RailStage]) -> str | None:
    legal_types = (int, float, bool, list, dict, str, None)

    for key, val in stage_class.config_options.items():
        def_val_dtype = None
        if isinstance(val, ceci.config.StageConfig):
            continue
        if isinstance(val, ceci.config.StageParameter):
            dtype = val.dtype
            if val.default is not None:
                def_val_dtype = type(val.default)
        elif isinstance(val, type):
            dtype = val
        else:
            dtype = type(val)
        if dtype not in legal_types:  # pragma: no cover
            return f"Illegal parameter type for {stage_class.name}.{key} {dtype}"
        if def_val_dtype not in legal_types:  # pragma: no cover
            return f"Illegal parameter default value type for {stage_class.name}.{key} {def_val_dtype}"
    return None


def build_and_read_pipeline(pipeline_class: str, **kwargs: Any) -> None:
    short_name = pipeline_class.split(".")[-1]
    yaml_file = f"{short_name}.yml"
    config_yaml_file = f"{short_name}_config.yml"
    build_pipeline(pipeline_class, yaml_file, "rubin", **kwargs)
    _pr = ceci.Pipeline.read(yaml_file)
    os.unlink(yaml_file)
    os.unlink(config_yaml_file)



"""Core code for RAIL"""

from .common_params import (SHARED_PARAMS, SharedParams, copy_param,
                            set_param_default, set_param_defaults)
from .data import (DATA_STORE, DataHandle, DataStore, FitsHandle, Hdf5Handle,
                   ModelDict, ModelHandle, PqHandle, QPDictHandle, QPHandle,
                   QPOrTableHandle, TableHandle)
from .introspection import RailEnv
from .model import Model
from .point_estimation import PointEstimationMixin
from .stage import RailPipeline, RailStage, StageIO

__all__ = [
    "SharedParams",
    "copy_param",
    "set_param_default",
    "set_param_defaults",
    "SHARED_PARAMS",
    "DataHandle",
    "TableHandle",
    "Hdf5Handle",
    "FitsHandle",
    "PqHandle",
    "QPHandle",
    "QPDictHandle",
    "QPOrTableHandle",
    "Model",
    "ModelDict",
    "ModelHandle",
    "DataStore",
    "DATA_STORE",
    "PointEstimationMixin",
    "StageIO",
    "RailPipeline",
    "RailStage",
    "RailEnv",
]


def find_version() -> str:
    """Find the version"""
    # setuptools_scm should install a
    # file _version alongside this one.
    from . import _version  # pylint: disable=import-outside-toplevel

    return _version.version


try:
    __version__ = find_version()
except ImportError:  # pragma: no cover
    __version__ = "unknown"



"""Abstract base classes defining a Creator.

A Creator will create synthetic photometric data and a PosteriorCalculator, 
which can calculate posteriors for the data with respect to the distribution 
defined by the Creator.
"""

from typing import Any

from ceci.config import StageParameter as Param

from rail.core.data import (DataHandle, ModelHandle, ModelLike, QPHandle,
                            TableHandle, TableLike)
from rail.core.stage import RailStage


class Modeler(RailStage):  # pragma: no cover
    """Base class for creating a model of redshift and photometry."""

    name = "Modeler"
    config_options = RailStage.config_options.copy()
    config_options.update(
        seed=Param(int, default=12345, msg="Random number seed"),
    )
    inputs = [("input", DataHandle)]
    outputs = [("model", ModelHandle)]

    def __init__(self, args: Any, **kwargs: Any):
        """Initialize Modeler"""
        super().__init__(args, **kwargs)
        self.model = None

    def fit_model(self) -> ModelHandle:
        """Produce a creation model from which photometry and redshifts can be
        generated.

        Returns
        -------
        ModelHandle:
            This will definitely be a wrapper around a File,
            but the filetype and format depend entirely on the
            modeling approach
        """
        self.validate()
        self.run()
        self.finalize()
        return self.get_handle("model")


class Creator(RailStage):  # pragma: no cover
    """Base class for Creators that generate synthetic photometric data from a
    model.

    ``Creator`` will output a table of photometric data. The details will depend
    on the particular engine.
    """

    name = "Creator"
    config_options = RailStage.config_options.copy()
    config_options.update(
        n_samples=Param(int, required=True, msg="Number of samples to create"),
        seed=Param(int, default=12345, msg="Random number seed"),
    )
    inputs = [("model", ModelHandle)]
    outputs = [("output", TableHandle)]

    def __init__(self, args: Any, **kwargs: Any) -> None:
        """Initialize Creator"""
        super().__init__(args, **kwargs)
        self.model = None
        if not isinstance(args, dict):  # pragma: no cover
            args = vars(args)
        self.open_model(**args)

    def sample(
        self, n_samples: int, seed: int | None = None, **kwargs: Any
    ) -> DataHandle:
        """Draw samples from the model specified in the configuration.

        This is a method for running a Creator in interactive mode. In pipeline
        mode, the subclass ``run`` method will be called by itself.

        Parameters
        ----------
        n_samples
            The number of samples to draw

        seed
            The random seed to control sampling

        **kwargs:
            Used to update the configuration

        Returns
        -------
        DataHandle
            DataHandle wrapping the newly created samples

        Notes
        -----
        This method puts ``n_samples`` and ``seed`` into the stage configuration
        data, which makes them available to other methods.

        It then calls the ``run`` method, which must be defined by a subclass.

        Finally, the ``DataHandle`` associated to the ``output`` tag is returned.
        """
        self.config["n_samples"] = n_samples
        self.config["seed"] = seed
        self.config.update(**kwargs)
        self.run()
        self.finalize()
        return self.get_handle("output")


class PosteriorCalculator(RailStage):  # pragma: no cover
    """Base class for object that calculates the posterior distribution of a
    particular field in a table of photometric data (typically the redshift).

    The posteriors will be contained in a qp Ensemble.
    """

    name = "PosteriorCalculator"
    config_options = RailStage.config_options.copy()
    config_options.update(
        column=Param(str, required=True, msg="Column to compute posterior for"),
    )
    inputs = [
        ("model", ModelHandle),
        ("input", TableHandle),
    ]
    outputs = [("output", QPHandle)]

    def __init__(self, args: Any, **kwargs: Any) -> None:
        """Initialize PosteriorCalculator"""
        super().__init__(args, **kwargs)
        self.model = None
        if not isinstance(args, dict):  # pragma: no cover
            args = vars(args)
        self.open_model(**args)

    def get_posterior(self, input_data: TableLike, **kwargs: Any) -> DataHandle:
        """Return posteriors for the given column.

        This is a method for running a Creator in interactive mode. In pipeline
        mode, the subclass ``run`` method will be called by itself.

        Parameters
        ----------
        data
            A table of the galaxies for which posteriors are calculated

        **kwargs
            Used to update configuration

        Returns
        -------
        DataHandle
            Posterior Estimate

        Notes
        -----
        This will put the ``data`` argument input this Stages the DataStore
        using this stages ``input`` tag.

        This will put the additional functional arguments into this Stages
        configuration data.

        It will then call ``self.run()`` and return the ``DataHandle``
        associated to the ``output`` tag.
        """
        self.set_data("input", input_data)
        self.config.update(**kwargs)
        self.run()
        self.finalize()
        return self.get_handle("output")



import os

import numpy as np
import qp

from rail.estimation.algos.equal_count import EqualCountClassifier
import rail.evaluation.metrics.pointestimates as pe
from rail.core.data import QPHandle, QPOrTableHandle, TableHandle
from rail.core.stage import RailStage
from rail.evaluation.dist_to_dist_evaluator import DistToDistEvaluator
from rail.evaluation.dist_to_point_evaluator import DistToPointEvaluator
from rail.evaluation.evaluator import OldEvaluator
from rail.evaluation.metrics.tomography import KDEBinOverlap
from rail.evaluation.point_to_point_evaluator import (
    PointToPointBinnedEvaluator, PointToPointEvaluator)
from rail.evaluation.single_evaluator import SingleEvaluator


# values for metrics
OUTRATE = 0.0
KSVAL = 0.367384
CVMVAL = 20.63155
ADVAL_ALL = 82.51480
ADVAL_CUT = 1.10750
CDEVAL = -4.31200
SIGIQR = 0.0045947
BIAS = -0.00001576
OUTRATE = 0.0
SIGMAD = 0.0046489


def construct_test_ensemble() -> tuple:
    np.random.seed(87)
    nmax = 2.5
    NPDF = 399
    true_zs = np.random.uniform(high=nmax, size=NPDF)
    locs = np.expand_dims(true_zs + np.random.normal(0.0, 0.01, NPDF), -1)
    true_ez = (locs.flatten() - true_zs) / (1.0 + true_zs)
    scales = np.ones((NPDF, 1)) * 0.1 + np.random.uniform(size=(NPDF, 1)) * 0.05
    n_ens = qp.Ensemble(
        qp.stats.norm, data=dict(loc=locs, scale=scales)  # pylint: disable=no-member
    )
    zgrid = np.linspace(0, nmax, 301)
    grid_ens = n_ens.convert_to(qp.interp_gen, xvals=zgrid)
    return zgrid, true_zs, grid_ens, true_ez


def test_point_metrics() -> None:
    zgrid, zspec, pdf_ens, true_ez = construct_test_ensemble()
    zb = pdf_ens.mode(grid=zgrid).flatten()

    ez = pe.PointStatsEz(zb, zspec).evaluate()
    assert np.allclose(ez, true_ez, atol=1.0e-2)
    # grid limits ez vals to ~10^-2 tol

    sig_iqr = pe.PointSigmaIQR(zb, zspec).evaluate()
    assert np.isclose(sig_iqr, SIGIQR)

    bias = pe.PointBias(zb, zspec).evaluate()
    assert np.isclose(bias, BIAS)

    out_rate = pe.PointOutlierRate(zb, zspec).evaluate()
    assert np.isclose(out_rate, OUTRATE)

    sig_mad = pe.PointSigmaMAD(zb, zspec).evaluate()
    assert np.isclose(sig_mad, SIGMAD)


def test_evaluation_stage() -> None:
    DS = RailStage.data_store
    _zgrid, zspec, pdf_ens, _true_ez = construct_test_ensemble()
    pdf = DS.add_data("pdf", pdf_ens, QPHandle)
    truth_table = dict(redshift=zspec)
    truth = DS.add_data("truth", truth_table, TableHandle)
    evaluator = OldEvaluator.make_stage(name="Eval", redshift_col="redshift")
    evaluator.evaluate(pdf, truth)

    os.remove(
        evaluator.get_output(evaluator.get_aliased_tag("output"), final_name=True)
    )


def test_dist_to_dist_evaluator(get_evaluation_files: tuple[str, str]) -> None:
    pdfs_file, _ztrue_file = get_evaluation_files
    assert pdfs_file

    DS = RailStage.data_store
    DS.__class__.allow_overwrite = True
    stage_dict = dict(
        # metrics=['cvm', 'ks', 'rmse', 'kld', 'ad'],
        metrics=["rmse"],
        _random_state=None,
    )

    ensemble = DS.read_file(key="pdfs_data", handle_class=QPHandle, path=pdfs_file)

    dtd_stage = DistToDistEvaluator.make_stage(name="dist_to_dist", **stage_dict)
    dtd_stage_single = DistToDistEvaluator.make_stage(
        name="dist_to_dist_single", force_exact=True, **stage_dict
    )

    _dtd_results = dtd_stage.evaluate(ensemble, ensemble)
    _dtd_results_single = dtd_stage_single.evaluate(ensemble, ensemble)

    for stage in [dtd_stage, dtd_stage_single]:
        os.remove(stage.get_output(stage.get_aliased_tag("output"), final_name=True))
        os.remove(stage.get_output(stage.get_aliased_tag("summary"), final_name=True))


def test_dist_to_point_evaluator(get_evaluation_files: tuple[str, str]) -> None:
    pdfs_file, ztrue_file = get_evaluation_files
    assert pdfs_file

    DS = RailStage.data_store
    DS.__class__.allow_overwrite = True
    stage_dict = dict(
        metrics=["cdeloss", "pit", "brier"],
        _random_state=None,
        metric_config={
            "brier": {"limits": (0, 3.1)},
        },
        limits=[0.0, 3.1],
    )

    ensemble = DS.read_file(key="pdfs_data", handle_class=QPHandle, path=pdfs_file)
    ztrue_data = DS.read_file("ztrue_data", TableHandle, ztrue_file)

    dtp_stage = DistToPointEvaluator.make_stage(name="dist_to_point", **stage_dict)
    dtp_stage_single = DistToPointEvaluator.make_stage(
        name="dist_to_point_single", force_exact=True, **stage_dict
    )

    _dtp_results = dtp_stage.evaluate(ensemble, ztrue_data)
    dtp_results_single = dtp_stage_single.evaluate(ensemble, ztrue_data)

    qp_dict = qp.read_dict(dtp_results_single["single_distribution_summary"].path)
    assert qp_dict["pit"].npdf == 1

    qp_dict_handle = dtp_results_single["single_distribution_summary"]
    qp_dict_handle.data = None
    qp_dict_2 = qp_dict_handle.read()
    assert qp_dict_2["pit"].npdf == 1

    qp_dict_handle.data = None

    with qp_dict_handle.open() as qp_dict_data:
        assert qp_dict_data

    for stage in [dtp_stage, dtp_stage_single]:
        os.remove(stage.get_output(stage.get_aliased_tag("output"), final_name=True))
        os.remove(stage.get_output(stage.get_aliased_tag("summary"), final_name=True))
        os.remove(
            stage.get_output(
                stage.get_aliased_tag("single_distribution_summary"), final_name=True
            )
        )


def test_point_to_point_evaluator(get_evaluation_files: tuple[str, str]) -> None:
    pdfs_file, ztrue_file = get_evaluation_files
    assert pdfs_file

    DS = RailStage.data_store
    DS.__class__.allow_overwrite = True
    stage_dict = dict(
        metrics=[
            "point_stats_ez",
            "point_stats_iqr",
            "point_bias",
            "point_outlier_rate",
            "point_stats_sigma_mad",
        ],
        _random_state=None,
        hdf5_groupname="photometry",
        point_estimate_key="zmode",
        chunk_size=10000,
    )

    ensemble = DS.read_file(key="pdfs_data", handle_class=QPHandle, path=pdfs_file)
    ztrue_data = DS.read_file("ztrue_data", TableHandle, ztrue_file)

    ptp_stage = PointToPointEvaluator.make_stage(name="point_to_point", **stage_dict)
    ptp_stage_single = PointToPointEvaluator.make_stage(
        name="point_to_point_single", force_exact=True, **stage_dict
    )

    _ptp_results = ptp_stage.evaluate(ensemble, ztrue_data)
    _ptp_results_single = ptp_stage_single.evaluate(ensemble, ztrue_data)

    for stage in [ptp_stage, ptp_stage_single]:
        os.remove(stage.get_output(stage.get_aliased_tag("output"), final_name=True))
        os.remove(stage.get_output(stage.get_aliased_tag("summary"), final_name=True))


def test_point_to_point_binning_evaluator(
    get_evaluation_files: tuple[str, str]
) -> None:
    pdfs_file, ztrue_file = get_evaluation_files
    assert pdfs_file

    DS = RailStage.data_store
    DS.__class__.allow_overwrite = True
    stage_dict = dict(
        metrics=[
            "point_stats_ez",
            "point_stats_iqr",
            "point_bias",
            "point_outlier_rate",
            "point_stats_sigma_mad",
        ],
        _random_state=None,
        hdf5_groupname="photometry",
        point_estimate_key="zmode",
        chunk_size=10000,
    )

    ensemble = DS.read_file(key="pdfs_data", handle_class=QPHandle, path=pdfs_file)
    ztrue_data = DS.read_file("ztrue_data", TableHandle, ztrue_file)

    ptp_stage_binning = PointToPointBinnedEvaluator.make_stage(
        name="point_to_point_binning", **stage_dict
    )

    _ptp_results = ptp_stage_binning.evaluate(ensemble, ztrue_data)

    for stage in [ptp_stage_binning]:
        os.remove(stage.get_output(stage.get_aliased_tag("output"), final_name=True))
        os.remove(stage.get_output(stage.get_aliased_tag("summary"), final_name=True))


def test_single_evaluator(get_evaluation_files: tuple[str, str]) -> None:
    pdfs_file, ztrue_file = get_evaluation_files
    assert pdfs_file
    DS = RailStage.data_store
    DS.__class__.allow_overwrite = True
    stage_dict = dict(
        metrics=[
            "cvm",
            "ks",
            "omega",
            "kld",
            "cdeloss",
            "point_stats_ez",
            "point_stats_iqr",
        ],
        _random_state=None,
        hdf5_groupname="photometry",
        point_estimates=["zmode"],
        truth_point_estimates=["redshift"],
        chunk_size=1000,
    )
    ensemble_d = DS.add_data("pdfs_data_2", None, QPOrTableHandle, path=pdfs_file)
    ztrue_data_d = DS.add_data("ztrue_data_2", None, QPOrTableHandle, path=ztrue_file)

    single_stage = SingleEvaluator.make_stage(name="single", **stage_dict)
    single_stage_single = SingleEvaluator.make_stage(
        name="single_single", force_exact=True, **stage_dict
    )

    _single_results = single_stage.evaluate(ensemble_d, ztrue_data_d)
    _single_results_single = single_stage_single.evaluate(ensemble_d, ztrue_data_d)

    for stage in [single_stage, single_stage_single]:
        os.remove(stage.get_output(stage.get_aliased_tag("output"), final_name=True))
        os.remove(stage.get_output(stage.get_aliased_tag("summary"), final_name=True))


def test_kde_overlap_evaluator(get_evaluation_files: tuple[str, str]) -> None:
    pdfs_file, ztrue_file = get_evaluation_files
    assert pdfs_file

    DS = RailStage.data_store
    DS.__class__.allow_overwrite = True

    ensemble = DS.read_file(key="pdfs_data", handle_class=QPHandle, path=pdfs_file)
    ztrue_data = DS.read_file("ztrue_data", TableHandle, ztrue_file)

    binner = EqualCountClassifier.make_stage()
    binindex = binner.classify(ensemble)

    kde_overlap = KDEBinOverlap.make_stage(
        hdf5_groupname="photometry", redshift_col="redshift"
    )

    _output = kde_overlap.evaluate(binindex, ztrue_data)



import os

import pytest

from rail.utils.path_utils import find_rail_file


@pytest.fixture(name="get_evaluation_files", scope="package")
def get_evaluation_files(request: pytest.FixtureRequest) -> tuple[str, str]:
    possible_local_file = "./examples_data/evaluation_data/data/output_fzboost.hdf5"
    if os.path.exists(possible_local_file):
        pdfs_file = os.path.abspath(possible_local_file)
    else:
        pdfs_file = "examples_data/evaluation_data/data/output_fzboost.hdf5"
        try:
            os.makedirs(os.path.dirname(pdfs_file))
        except FileExistsError:
            pass
        curl_com = f"curl -o {pdfs_file} https://portal.nersc.gov/cfs/lsst/PZ/output_fzboost.hdf5"
        os.system(curl_com)
    ztrue_file = find_rail_file("examples_data/testdata/test_dc2_validation_9816.hdf5")

    def cleanup_evaluation_files() -> None:
        os.unlink(possible_local_file)

    request.addfinalizer(cleanup_evaluation_files)

    return pdfs_file, ztrue_file



"""
A summarizer that simple makes a histogram of a point estimate
"""

from typing import Any, Generator

import numpy as np
import qp
from ceci.config import StageParameter as Param

from rail.core.data import DataHandle, QPHandle, TableHandle, TableLike
from rail.estimation.informer import PzInformer
from rail.estimation.summarizer import PZSummarizer


class NaiveStackInformer(PzInformer):
    """Placeholder Informer"""

    name = "NaiveStackInformer"
    config_options = PzInformer.config_options.copy()

    def _finalize_run(self) -> None:
        self.model = np.array([None])
        PzInformer._finalize_run(self)


class NaiveStackSummarizer(PZSummarizer):
    """Summarizer which stacks individual P(z)"""

    name = "NaiveStackSummarizer"
    config_options = PZSummarizer.config_options.copy()
    config_options.update(
        zmin=Param(float, 0.0, msg="The minimum redshift of the z grid"),
        zmax=Param(float, 3.0, msg="The maximum redshift of the z grid"),
        nzbins=Param(int, 301, msg="The number of gridpoints in the z grid"),
        seed=Param(int, 87, msg="random seed"),
        nsamples=Param(int, 1000, msg="Number of sample distributions to create"),
    )
    inputs = [("input", QPHandle)]
    outputs = [("output", QPHandle), ("single_NZ", QPHandle)]

    def __init__(self, args: Any, **kwargs: Any) -> None:
        super().__init__(args, **kwargs)
        self.zgrid: np.ndarray | None = None

    def _setup_iterator(self) -> Generator:
        itr = self.input_iterator("input")
        for s, e, d in itr:
            yield s, e, d, np.ones(e - s, dtype=bool)

    def run(self) -> None:
        handle = self.get_handle("input", allow_missing=True)
        self._input_length = handle.size()
        iterator = self._setup_iterator()
        self.zgrid = np.linspace(
            self.config.zmin, self.config.zmax, self.config.nzbins + 1
        )
        assert self.zgrid is not None
        # Initiallizing the stacking pdf's
        yvals = np.zeros((1, len(self.zgrid)))
        bvals = np.zeros((self.config.nsamples, len(self.zgrid)))
        bootstrap_matrix = self._broadcast_bootstrap_matrix()

        first = True
        for s, e, test_data, mask in iterator:
            print(f"Process {self.rank} running estimator on chunk {s:,} - {e:,}")
            self._process_chunk(
                s, e, test_data, mask, first, bootstrap_matrix, yvals, bvals
            )
            first = False
        if self.comm is not None:  # pragma: no cover
            bvals, yvals = self._join_histograms(bvals, yvals)

        if self.rank == 0:
            sample_ens = qp.Ensemble(
                qp.interp, data=dict(xvals=self.zgrid, yvals=bvals)
            )
            qp_d = qp.Ensemble(qp.interp, data=dict(xvals=self.zgrid, yvals=yvals))
            self.add_data("output", sample_ens)
            self.add_data("single_NZ", qp_d)

    def _process_chunk(
        self,
        start: int,
        end: int,
        data: qp.Ensemble,
        mask: np.ndarray,
        _first: bool,
        bootstrap_matrix: np.ndarray,
        yvals: np.ndarray,
        bvals: np.ndarray,
    ) -> None:
        assert self.zgrid is not None
        pdf_vals = data.pdf(self.zgrid)
        squeeze_mask = np.squeeze(mask)
        yvals += np.expand_dims(
            np.sum(
                np.where(
                    np.isfinite(pdf_vals[squeeze_mask, :]), pdf_vals[squeeze_mask], 0.0
                ),
                axis=0,
            ),
            0,
        )
        # qp_d is the normalized probability of the stack, we need to know how many galaxies were
        for i in range(self.config.nsamples):
            bootstrap_draws = bootstrap_matrix[:, i]
            # Neither all of the bootstrap_draws are in this chunk nor the index starts at "start"
            chunk_mask = (bootstrap_draws >= start) & (bootstrap_draws < end)
            bootstrap_draws = bootstrap_draws[chunk_mask] - start
            zarr = np.where(squeeze_mask, pdf_vals.T, 0.0).T[bootstrap_draws]
            bvals[i] += np.sum(zarr, axis=0)


class NaiveStackMaskedSummarizer(NaiveStackSummarizer):
    name = "NaiveStackMaskedSummarizer"
    config_options = NaiveStackSummarizer.config_options.copy()
    config_options.update(
        selected_bin=Param(int, -1, msg="bin to use"),
    )
    inputs = [("input", QPHandle), ("tomography_bins", TableHandle)]
    outputs = [("output", QPHandle), ("single_NZ", QPHandle)]

    def _setup_iterator(self) -> Generator:
        selected_bin = self.config.selected_bin
        if self.config.tomography_bins in ["none", None]:
            selected_bin = -1

        if selected_bin == -1:
            itrs = [self.input_iterator("input")]
        else:
            itrs = [
                self.input_iterator("input"),
                self.input_iterator("tomography_bins"),
            ]

        for it in zip(*itrs):
            first = True
            mask = None
            for s, e, d in it:
                if first:
                    start = s
                    end = e
                    pz_data = d
                    first = False
                else:
                    mask = d["class_id"] == self.config.selected_bin
            if mask is None:
                mask = np.ones(pz_data.npdf, dtype=bool)
            yield start, end, pz_data, mask

    def summarize(
        self, input_data: qp.Ensemble, tomo_bins: TableLike | None = None
    ) -> DataHandle:
        """Override the Summarizer.summarize() method to take tomo bins
        as an additional input

        Parameters
        ----------
        input_data
            Per-galaxy p(z), and any ancilary data associated with it

        tomo_bins
            Tomographic bins file

        Returns
        -------
        DataHandle
            Ensemble with n(z), and any ancilary data
        """
        self.set_data("input", input_data)
        if tomo_bins is None:
            self.config.tomography_bins = None
        else:
            self.set_data("tomography_bins", tomo_bins)
        self.run()
        self.finalize()
        return self.get_handle("output")



"""Rail-specific data management"""

from __future__ import annotations

import enum
import os
import pickle
from typing import Any, Callable, Iterable, TypeAlias, TypeVar

import qp
import tables_io
from tables_io import hdf5 as tab_hdf5

from .model import Model

T = TypeVar("T", bound="DataHandle")

# These are place-holders for if and when we enforce typing
DataLike: TypeAlias = Any
GroupLike: TypeAlias = Any
ModelLike: TypeAlias = Any
TableLike: TypeAlias = Any
FileLike: TypeAlias = Any


class DataHandle:  # pylint: disable=too-many-instance-attributes
    """Class to act as a handle for a bit of data.  Associating it with a file and
    providing tools to read & write it to that file

    """

    suffix: str | None = ""

    # This is to keep track of all the sub-types
    _data_handle_type_dict: dict[str, type[DataHandle]] = {}

    def __init__(
        self,
        tag: str,
        data: DataLike | None = None,
        path: str | None = None,
        creator: str | None = None,
    ) -> None:
        """Constructor

        Parameters
        ----------
        tag
            The tag under which this data handle can be found in the store

        data
            The associated data

        path
            The path to the associated file

        creator
            The name of the stage that created this data handle
        """
        self.tag = tag
        if data is not None:
            self._validate_data(data)
        self.data = data
        self.path = path
        self.creator = creator
        self.fileObj: FileLike = None
        self.groups: GroupLike = None
        self.partial: bool | None = False
        self.length: int | None = None

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Register the subclass with the dict"""
        cls._data_handle_type_dict[cls.__name__] = cls

    @classmethod
    def get_sub_classes(cls) -> dict[str, type[DataHandle]]:
        """Get all the subclasses"""
        return cls._data_handle_type_dict

    @classmethod
    def get_sub_class(cls, class_name: str) -> type[DataHandle]:
        """Get a particular subclass by name"""
        return cls._data_handle_type_dict[class_name]

    @classmethod
    def print_sub_classes(cls) -> None:
        """Print the list of all the subclasses"""
        for key, val in cls._data_handle_type_dict.items():
            print(f"{key}: {val}")

    def open(self, **kwargs: Any) -> FileLike:
        """Open and return the associated file

        Parameters
        ----------
        **kwargs
            Passed to the call to open the file in question

        Returns
        -------
        FileLike
            Newly opened file

        Notes
        -----
        This will simply open the file and return a FileLike object to the caller.
        It will not read or cache the data
        """
        if self.path is None:
            raise ValueError("DataHandle.open() called but path has not been specified")
        self.fileObj = self._open(os.path.expandvars(self.path), **kwargs)
        return self.fileObj

    @classmethod
    def _open(cls, path: str, **kwargs: Any) -> FileLike:
        raise NotImplementedError("DataHandle._open")  # pragma: no cover

    def close(self, **kwargs: Any) -> None:  # pylint: disable=unused-argument
        """Close the associated file"""
        self.fileObj = None

    def read(self, force: bool = False, **kwargs: Any) -> DataLike:
        """Read and return the data from the associated file

        Parameters
        ----------
        force
            If true, force re-reading the data

        **kwargs
            Passed to the call to read the data

        Returns
        -------
        DataLike
            Data that were read

        Notes
        -----
        This will read the entire file, and while useful for testing on small files,
        will not work on very large files.
        """
        if self.data is not None and not force:
            return self.data
        assert self.path is not None
        self.set_data(self._read(os.path.expandvars(self.path), **kwargs))
        return self.data

    def __call__(self, **kwargs: Any) -> DataLike:
        """Return the data, re-reading the fill if needed"""
        if self.has_data and not self.partial:
            return self.data
        return self.read(force=True, **kwargs)

    @classmethod
    def _read(cls, path: str, **kwargs: Any) -> DataLike:
        raise NotImplementedError("DataHandle._read")  # pragma: no cover

    def write(self, **kwargs: Any) -> None:
        """Write the data to the associated file"""
        if self.path is None:
            raise ValueError(
                "TableHandle.write() called but path has not been specified"
            )
        if self.data is None:
            raise ValueError(
                f"TableHandle.write() called for path {self.path} with no data"
            )
        outdir = os.path.dirname(os.path.abspath(os.path.expandvars(self.path)))
        if not os.path.exists(outdir):  # pragma: no cover
            os.makedirs(outdir, exist_ok=True)
        return self._write(self.data, os.path.expandvars(self.path), **kwargs)

    @classmethod
    def _write(cls, data: DataLike, path: str, **kwargs: Any) -> None:
        raise NotImplementedError("DataHandle._write")  # pragma: no cover

    def initialize_write(self, data_length: int, **kwargs: Any) -> None:
        """Initialize file to be written by chunks

        Parameters
        ----------
        data_length
            Number of rows of data that we will write, used to reserve space

        **kwargs
            Information about the columns we will write
        """
        if self.path is None:  # pragma: no cover
            raise ValueError(
                "TableHandle.write() called but path has not been specified"
            )
        self.groups, self.fileObj = self._initialize_write(
            self.data, os.path.expandvars(self.path), data_length, **kwargs
        )

    @classmethod
    def _initialize_write(
        cls,
        data: DataLike,
        path: str,
        data_length: int,
        **kwargs: Any,
    ) -> tuple[GroupLike, FileLike]:
        raise NotImplementedError("DataHandle._initialize_write")  # pragma: no cover

    def write_chunk(self, start: int, end: int, **kwargs: Any) -> None:
        """Write the data to the associated file

        Parameters
        ----------
        start
            Index of starting row for this chunk of data

        end
            Index of ending row for this chunk of data

        **kwargs
            Passed to call to write this chunk of data
        """
        if self.data is None:
            raise ValueError(
                f"TableHandle.write_chunk() called for path {self.path} with no data"
            )
        if self.fileObj is None:
            raise ValueError(
                f"TableHandle.write_chunk() called before open for {self.tag} : {self.path}"
            )
        return self._write_chunk(
            self.data, self.fileObj, self.groups, start, end, **kwargs
        )

    @classmethod
    def _write_chunk(
        cls,
        data: DataLike,
        fileObj: FileLike,
        groups: GroupLike,
        start: int,
        end: int,
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError("DataHandle._write_chunk")  # pragma: no cover

    def finalize_write(self, **kwargs: Any) -> None:
        """Finalize and close file written by chunks

        Parameters
        ----------
        **kwargs
            Passed to call to write this chunk of data
        """
        if self.fileObj is None:  # pragma: no cover
            raise ValueError(
                f"TableHandle.finalize_wite() called before open for {self.tag} : {self.path}"
            )
        self._finalize_write(self.data, self.fileObj, **kwargs)

    @classmethod
    def _finalize_write(cls, data: DataLike, fileObj: FileLike, **kwargs: Any) -> None:
        raise NotImplementedError("DataHandle._finalize_write")  # pragma: no cover

    def iterator(self, **kwargs: Any) -> Iterable:
        """Iterator over the data"""
        if self.data is not None and self.partial is False:
            return self._in_memory_iterator(**kwargs)
        assert self.path is not None
        return self._iterator(self.path, **kwargs)

    def set_data(self, data: DataLike, partial: bool = False) -> None:
        """Set the data for a chunk, and set the partial flag to true"""
        self._validate_data(data)
        self.data = data
        self.partial = partial

    @classmethod
    def _validate_data(cls, data: DataLike) -> None:  # pylint: disable=unused-argument
        """Make sure that the right type of data is being passed in"""
        return

    def size(self, **kwargs: Any) -> int:
        """Return the size of the data associated to this handle"""
        if self.partial:
            if self.length is not None:  # pragma: no cover
                return self.length
            assert self.path is not None
            return self._size(self.path, **kwargs)
        if self.data is not None:
            return self.data_size(**kwargs)
        assert self.path is not None
        return self._size(self.path, **kwargs)

    def _size(self, path: str, **kwargs: Any) -> int:
        raise NotImplementedError("DataHandle._size")  # pragma: no cover

    def data_size(self, **kwargs: Any) -> int:
        """Return the size of the in memorry data"""
        if self.data is None:  # pragma: no cover
            return 0
        return self._data_size(self.data, **kwargs)

    def _data_size(self, data: DataLike, **kwargs: Any) -> int:
        raise NotImplementedError("DataHandle._data_size")  # pragma: no cover

    def _in_memory_iterator(self, **kwargs: Any) -> Iterable:
        raise NotImplementedError("DataHandle._in_memory_iterator")  # pragma: no cover

    @classmethod
    def _iterator(cls, path: str, **kwargs: Any) -> Iterable:
        raise NotImplementedError("DataHandle._iterator")  # pragma: no cover

    @property
    def has_data(self) -> bool:
        """Return true if the data for this handle are loaded"""
        return self.data is not None

    @property
    def has_path(self) -> bool:
        """Return true if the path for the associated file is defined"""
        return self.path is not None

    @property
    def is_written(self) -> bool:
        """Return true if the associated file has been written"""
        if self.path is None:
            return False
        return os.path.exists(os.path.expandvars(self.path))

    def __str__(self) -> str:
        s = f"{type(self)} "
        if self.has_path:
            s += f"{self.path}, ("
        else:
            s += "None, ("
        if self.is_written:
            s += "w"
        if self.has_data:
            s += "d"
        s += ")"
        return s

    @classmethod
    def make_name(cls, tag: str) -> str:
        """Construct and return file name for a particular data tag"""
        if cls.suffix:
            return f"{tag}.{cls.suffix}"
        return tag  # pragma: no cover

    @classmethod
    def _check_data_columns(
        cls,
        path: str,
        columns_to_check: list[str],
        parent_groupname: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Checking if certain columns required by the stage is present in the data"""
        # path: the path to data file
        # columns_to_check: list of columns required by the specific stage
        # kwargs: other key arguments required by specific data type, see below
        raise NotImplementedError  # pragma: no cover


class TableHandle(DataHandle):
    """DataHandle for single tables of data"""

    suffix: str | None = None

    def set_data(self, data: TableLike, partial: bool = False) -> None:
        """Set the data for a chunk, and set the partial flag if this is not all the data"""
        self._validate_data(data)
        self.data = data
        self.partial = partial

    @classmethod
    def _validate_data(cls, data: TableLike) -> None:  # pylint: disable=unused-argument
        """Make sure that the right type of data is being passed in"""
        return

    @classmethod
    def _open(cls, path: str, **kwargs: Any) -> FileLike:
        """Open and return the associated file

        Notes
        -----
        This will simply open the file and return a file-like object to the caller.
        It will not read or cache the data
        """
        return tables_io.io_open(path, **kwargs)  # pylint: disable=no-member

    @classmethod
    def _read(cls, path: str, **kwargs: Any) -> TableLike:
        """Read and return the data from the associated file"""
        return tables_io.read(path, **kwargs)

    @classmethod
    def _write(cls, data: TableLike, path: str, **kwargs: Any) -> None:
        """Write the data to the associated file"""
        return tables_io.write(data, path, **kwargs)

    def _size(self, path: str, **kwargs: Any) -> int:
        if path in [None, "none", "None"]:  # pragma: no cover
            return 0
        return tab_hdf5.get_input_data_length(path, **kwargs)

    def _data_size(self, data: TableLike, **kwargs: Any) -> int:
        group_name = kwargs.get("groupname", None)
        if group_name:
            try:
                data = data[group_name]
            except KeyError:  # pragma: no cover
                pass
        max_l = 0
        for _k, v in data.items():
            max_l = max(max_l, len(v))
        return max_l

    def _in_memory_iterator(self, **kwargs: Any) -> Iterable:
        nrows = self.data_size(**kwargs)
        groupname = kwargs.get("groupname", None)
        if isinstance(self.data, dict) and groupname:
            try:
                table = self.data[groupname]
            except KeyError:  # pragma: no cover
                table = self.data
        else:
            table = self.data
        for start, end in tab_hdf5.data_ranges_by_rank(
            nrows, kwargs.get("chunk_size", 100000), 1, 0
        ):
            if isinstance(self.data, dict):
                yield start, end, tables_io.array_utils.slice_dict(
                    table, slice(start, end)
                )
            else:  # pragma: no cover
                yield start, end, table[start:end]

    @classmethod
    def _iterator(cls, path: str, **kwargs: Any) -> Iterable:
        """Iterate over the data"""
        return tables_io.iteratorNative(path, **kwargs)

    @classmethod
    def _check_data_columns(
        cls,
        path: str,
        columns_to_check: list[str],
        parent_groupname: str | None = None,
        **kwargs: Any,
    ) -> None:
        tables_io.check_columns(
            path, columns_to_check, parent_groupname=parent_groupname, **kwargs
        )


class Hdf5Handle(TableHandle):  # pragma: no cover
    """DataHandle for a table written to HDF5"""

    suffix = "hdf5"

    @classmethod
    def _initialize_write(
        cls, data: TableLike, path: str, data_length: int, **kwargs: Any
    ) -> tuple[GroupLike, FileLike]:
        initial_dict = cls._get_allocation_kwds(data, data_length)
        comm = kwargs.get("communicator", None)
        group, fout = tab_hdf5.initialize_HDF5_write_single(
            path, groupname=None, comm=comm, **initial_dict
        )
        return group, fout

    @classmethod
    def _get_allocation_kwds(cls, data: TableLike, data_length: int) -> dict:
        keywords = {}
        for key, array in data.items():
            shape = list(array.shape)
            if not shape:
                continue
            shape[0] = data_length
            keywords[key] = (shape, array.dtype)
        return keywords

    @classmethod
    def _write_chunk(
        cls,
        data: TableLike,
        fileObj: FileLike,
        groups: GroupLike,
        start: int,
        end: int,
        **kwargs: Any,
    ) -> None:
        tab_hdf5.write_dict_to_HDF5_chunk_single(fileObj, data, start, end, **kwargs)

    @classmethod
    def _finalize_write(cls, data: TableLike, fileObj: FileLike, **kwargs: Any) -> None:
        return tab_hdf5.finalize_HDF5_write(fileObj, **kwargs)


class FitsHandle(TableHandle):
    """DataHandle for a table written to fits"""

    suffix = "fits"


class PqHandle(TableHandle):
    """DataHandle for a parquet table"""

    suffix = "pq"

    def _size(self, path: str, **kwargs: Any) -> int:
        return tab_hdf5.get_input_data_length(path, **kwargs)


class QPHandle(DataHandle):
    """DataHandle for qp ensembles"""

    suffix = "hdf5"

    @classmethod
    def _open(cls, path: str, **kwargs: Any) -> FileLike:
        """Open and return the associated file

        Notes
        -----
        This will simply open the file and return a FileLike object to the caller.
        It will not read or cache the data
        """
        return tables_io.io_open(path, **kwargs)  # pylint: disable=no-member

    @classmethod
    def _read(cls, path: str, **kwargs: Any) -> qp.Ensemble:
        """Read and return the data from the associated file"""
        return qp.read(path)

    @classmethod
    def _write(cls, data: qp.Ensemble, path: str, **kwargs: Any) -> None:
        """Write the data to the associated file"""
        return data.write_to(path)

    @classmethod
    def _initialize_write(
        cls, data: qp.Ensemble, path: str, data_length: int, **kwargs: Any
    ) -> tuple[GroupLike, FileLike]:
        comm = kwargs.get("communicator", None)
        return data.initializeHdf5Write(path, data_length, comm)

    @classmethod
    def _write_chunk(
        cls,
        data: qp.Ensemble,
        fileObj: FileLike,
        groups: GroupLike,
        start: int,
        end: int,
        **kwargs: Any,
    ) -> None:
        return data.writeHdf5Chunk(fileObj, start, end)

    @classmethod
    def _finalize_write(
        cls, data: qp.Ensemble, fileObj: FileLike, **kwargs: Any
    ) -> None:
        return data.finalizeHdf5Write(fileObj)

    @classmethod
    def _validate_data(cls, data: qp.Ensemble) -> None:
        if not isinstance(data, qp.Ensemble):
            raise TypeError(
                f"Expected `data` to be a `qp.Ensemble`, but {type(data)} was provided."
                "Perhaps you meant to use `TableHandle`?"
            )

    def _size(self, path: str, **kwargs: Any) -> int:
        if self.data is not None and not self.partial:  # pragma: no cover
            return self._data_size(self.data)
        if path in [None, "none", "None"]:  # pragma: no cover
            return 0
        return tab_hdf5.get_input_data_length(path, groupname="data")

    def _data_size(self, data: qp.Ensemble, **kwargs: Any) -> int:
        return self.data.npdf

    def _in_memory_iterator(self, **kwargs: Any) -> Iterable:
        nrows = self.data.npdf
        for start, end in tab_hdf5.data_ranges_by_rank(
            nrows, kwargs.get("chunk_size", 100000), 1, 0
        ):
            yield start, end, self.data[start:end]

    @classmethod
    def _iterator(cls, path: str, **kwargs: Any) -> Iterable:
        """Iterate over the data"""
        kwargs.pop("groupname", "None")
        return qp.iterator(path, **kwargs)


class QPDictHandle(DataHandle):
    """DataHandle for dictionaries of qp ensembles"""

    suffix = "hdf5"

    @classmethod
    def _open(cls, path: str, **kwargs: Any) -> FileLike:
        """Open and return the associated file

        Notes
        -----
        This will simply open the file and return a FileLike object to the caller.
        It will not read or cache the data
        """
        return tables_io.io_open(path, **kwargs)  # pylint: disable=no-member

    @classmethod
    def _read(cls, path: str, **kwargs: Any) -> dict[str, qp.Ensemble]:
        """Read and return the dictionary of qp.Ensembles from the associated file"""
        return qp.read_dict(path)

    @classmethod
    def _write(cls, data: dict[str, qp.Ensemble], path: str, **kwargs: Any) -> None:
        """Write the data (a dictionary of qp.Ensembles) to the associated file"""
        return qp.write_dict(path, data)


class QPOrTableHandle(QPHandle, Hdf5Handle):
    """DataHandle that will work with either qp.Ensembles or TableLike data"""

    suffix = "hdf5"

    class PdfOrValue(enum.Enum):
        unknown = -1
        distribution = 0
        point_estimate = 1
        both = 2

        def has_dist(self) -> bool:
            return self.value in [0, 2]

        def has_point(self) -> bool:
            return self.value in [1, 2]

    def is_qp(self) -> bool:
        """Check if the associated data or file is a QP ensemble"""
        if self.path in [None, "None", "none"]:
            return isinstance(self.data, qp.Ensemble)
        return qp.is_qp_file(self.path)

    def check_pdf_or_point(self) -> PdfOrValue:
        """Check the associated file to see if it is a QP pdf, point estimate or both"""
        if self.is_qp():
            return self.PdfOrValue.both
        return self.PdfOrValue.point_estimate

    @classmethod
    def _open(cls, path: str, **kwargs: Any) -> FileLike:
        """Open and return the associated file

        Notes
        -----
        This will simply open the file and return a file-like object to the caller.
        It will not read or cache the data
        """
        return tables_io.io_open(path, **kwargs)  # pylint: disable=no-member

    @classmethod
    def _read(cls, path: str, **kwargs: Any) -> qp.Ensemble | TableLike:
        """Read and return the data from the associated file"""
        if qp.is_qp_file(path):
            return qp.read(path, **kwargs)
        return tables_io.read(path, **kwargs)

    @classmethod
    def _write(cls, data: qp.Ensemble | TableLike, path: str, **kwargs: Any) -> None:
        """Write the data to the associated file"""
        raise RuntimeError(
            "QPOrTableHandle should be used for input, not output"
        )  # pragma: no cover

    @classmethod
    def _initialize_write(
        cls, data: qp.Ensemble | TableLike, path: str, data_length: int, **kwargs: Any
    ) -> tuple[GroupLike, FileLike]:
        raise RuntimeError(
            "QPOrTableHandle should be used for input, not output"
        )  # pragma: no cover

    @classmethod
    def _write_chunk(
        cls,
        data: qp.Ensemble | TableLike,
        fileObj: FileLike,
        groups: GroupLike,
        start: int,
        end: int,
        **kwargs: Any,
    ) -> None:
        raise RuntimeError(
            "QPOrTableHandle should be used for input, not output"
        )  # pragma: no cover

    @classmethod
    def _finalize_write(
        cls, data: qp.Ensemble | TableLike, fileObj: FileLike, **kwargs: Any
    ) -> None:
        raise RuntimeError(
            "QPOrTableHandle should be used for input, not output"
        )  # pragma: no cover

    @classmethod
    def _validate_data(cls, data: qp.Ensemble | TableLike) -> None:
        pass

    def _size(self, path: str, **kwargs: Any) -> int:
        if self.is_qp():
            return QPHandle._size(self, path, **kwargs)
        return Hdf5Handle._size(self, path, **kwargs)

    def _data_size(self, data: qp.Ensemble | TableLike, **kwargs: Any) -> int:
        if self.is_qp():
            return self.data.npdf
        return Hdf5Handle._data_size(self, data, **kwargs)

    def _in_memory_iterator(self, **kwargs: Any) -> Iterable:
        if self.is_qp():
            return QPHandle._in_memory_iterator(self, **kwargs)
        return Hdf5Handle._in_memory_iterator(self, **kwargs)

    @classmethod
    def _iterator(cls, path: str, **kwargs: Any) -> Iterable:
        """Iterate over the data"""
        if qp.is_qp_file(path):
            return QPHandle._iterator(path, **kwargs)
        return Hdf5Handle._iterator(path, **kwargs)


def default_model_read(modelfile: str) -> ModelLike:
    """Default function to read model files, simply used pickle.load"""
    with open(modelfile, "rb") as fin:
        read_data = pickle.load(fin)
    if isinstance(read_data, Model):
        ret_data = read_data.data
    else:
        ret_data = read_data
    return ret_data


def default_model_write(model: ModelLike, path: str) -> None:
    """Write the model, this default implementation uses pickle"""
    with open(path, "wb") as fout:
        pickle.dump(obj=model, file=fout, protocol=pickle.HIGHEST_PROTOCOL)


class ModelDict(dict):
    """
    A specialized dict to keep track of individual estimation models objects:
    this is just a dict these additional features

    1. Keys are paths
    2. There is a read(path, force=False) method that reads a model object and
        inserts it into the dictionary
    3. There is a single static instance of this class

    """

    def open(self, path: str, mode: str, **kwargs: Any) -> FileLike:
        """Open the file and return the file handle"""
        return open(path, mode, **kwargs)  # pylint: disable=unspecified-encoding

    def read(
        self,
        path: str,
        force: bool = False,
        reader: Callable | None = None,
        **_kwargs: Any,
    ) -> ModelLike:  # pylint: disable=unused-argument
        """Read a model into this dict"""
        if reader is None:
            reader = default_model_read
        if force or path not in self:
            model = reader(path)
            self[path] = model
            return model
        return self[path]

    def write(
        self,
        model: ModelLike,
        path: str,
        force: bool = False,
        writer: Callable | None = None,
        **_kwargs: Any,
    ) -> None:  # pylint: disable=unused-argument
        """Write the model, this default implementation uses pickle"""
        if writer is None:
            writer = default_model_write
        if force or path not in self or not os.path.exists(path):
            self[path] = model
            writer(model, path)


class ModelHandle(DataHandle):
    """DataHandle for machine learning models"""

    suffix = "pkl"

    model_factory = ModelDict()

    @classmethod
    def _open(cls, path: str, **kwargs: Any) -> FileLike:
        """Open and return the associated file"""
        kwcopy = kwargs.copy()
        if kwcopy.pop("mode", "r") == "w":
            return cls.model_factory.open(path, mode="wb", **kwcopy)
        force = kwargs.pop("force", False)
        reader = kwargs.pop("reader", None)
        assert isinstance(force, bool)
        return cls.model_factory.read(path, force=force, reader=reader, **kwargs)

    @classmethod
    def _read(cls, path: str, **kwargs: Any) -> ModelLike:
        """Read and return the data from the associated file"""
        force = kwargs.pop("force", False)
        reader = kwargs.pop("reader", None)
        assert isinstance(force, bool)
        return cls.model_factory.read(path, force=force, reader=reader, **kwargs)

    @classmethod
    def _write(cls, data: ModelLike, path: str, **kwargs: Any) -> None:
        """Write the data to the associated file"""
        force = kwargs.pop("force", False)
        writer = kwargs.pop("writer", None)
        assert isinstance(force, bool)
        return cls.model_factory.write(data, path, force=force, writer=writer, **kwargs)


class DataStore(dict):
    """Class to provide a transient data store

    This class:

    1. associates data products with keys
    2. provides functions to read and write the various data produces to associated files

    """

    allow_overwrite = False

    def __init__(self, **kwargs: Any) -> None:
        """Build from keywords

        Note
        ----
        All of the values must be data handles of this will raise a TypeError
        """
        super(dict, self).__init__()
        for key, val in kwargs.items():
            self[key] = val

    def __str__(self) -> str:
        """Override __str__ casting to deal with `TableHandle` objects in the map"""
        s = "{"
        for key, val in self.items():
            s += f"  {key}:{val}\n"
        s += "}"
        return s

    def __repr__(self) -> str:
        """A custom representation"""
        s = "DataStore\n"
        s += self.__str__()
        return s

    def __setitem__(self, key: str, value: DataHandle) -> None:
        """Override the __setitem__ to work with ``TableHandle``"""
        if not isinstance(value, DataHandle):
            raise TypeError(
                f"Can only add objects of type DataHandle to DataStore, not {type(value)}"
            )
        check = self.get(key)
        if check is not None and not self.allow_overwrite:
            raise ValueError(
                f"DataStore already has an item with key {key},"
                f"of type {type(check)}, created by {check.creator}"
            )
        dict.__setitem__(self, key, value)

    def __getattr__(self, key: str) -> DataHandle | None:
        """Allow attribute-like parameter access"""
        try:
            return self.__getitem__(key)
        except KeyError as msg:
            # Kludge to get docstrings to work
            if key in ["__objclass__"]:  # pragma: no cover
                return None
            raise KeyError from msg

    def __setattr__(self, key: str, value: DataHandle) -> None:
        """Allow attribute-like parameter setting"""
        self.__setitem__(key, value)

    def add_data(
        self,
        key: str,
        data: DataLike,
        handle_class: type[DataHandle],
        path: str | None = None,
        creator: str = "DataStore",
    ) -> DataHandle:
        """Create a handle for some data, and insert it into the DataStore"""
        handle = handle_class(key, path=path, data=data, creator=creator)
        self[key] = handle
        return handle

    def add_handle(
        self,
        key: str,
        handle_class: type[DataHandle],
        path: str,
        creator: str = "DataStore",
    ) -> DataHandle:
        """Create a handle for some data, and insert it into the DataStore"""
        handle = handle_class(key, path=path, data=None, creator=creator)
        self[key] = handle
        return handle

    def read_file(
        self,
        key: str,
        handle_class: type[DataHandle],
        path: str,
        creator: str = "DataStore",
        **kwargs: Any,
    ) -> DataHandle:
        """Create a handle, use it to read a file, and insert it into the DataStore"""
        handle = handle_class(key, path=path, data=None, creator=creator)
        handle.read(**kwargs)
        self[key] = handle
        return handle

    def read(self, key: str, force: bool = False, **kwargs: Any) -> DataLike:
        """Read the data associated to a particular key"""
        try:
            handle = self[key]
        except KeyError as msg:
            raise KeyError(f"Failed to read data {key} because {msg}") from msg
        return handle.read(force, **kwargs)

    def open(self, key: str, mode: str = "r", **kwargs: Any) -> FileLike:
        """Open and return the file associated to a particular key"""
        try:
            handle = self[key]
        except KeyError as msg:
            raise KeyError(f"Failed to open data {key} because {msg}") from msg
        return handle.open(mode=mode, **kwargs)

    def write(self, key: str, **kwargs: Any) -> None:
        """Write the data associated to a particular key"""
        try:
            handle = self[key]
        except KeyError as msg:
            raise KeyError(f"Failed to write data {key} because {msg}") from msg
        return handle.write(**kwargs)

    def write_all(self, force: bool = False, **kwargs: Any) -> None:
        """Write all the data in this DataStore"""
        for key, handle in self.items():
            local_kwargs = kwargs.get(key, {})
            if handle.is_written and not force:
                continue
            handle.write(**local_kwargs)


_DATA_STORE = DataStore()


def DATA_STORE() -> DataStore:
    """Return the factory instance"""
    return _DATA_STORE



import importlib
import os
import pkgutil
from types import ModuleType

import setuptools

import rail


class RailEnv:
    """Singleton class that manages introspection into the RAIL installation

    This will idenfity all of the installed RAIL packages, namespaces and
    files, and navigate the connections between them.

    This will also generate restructed text (`.rst`) files for the Sphinx
    autodoc documentation
    """

    _packages: dict[str, pkgutil.ModuleInfo] = {}
    _namespace_path_dict: dict[str, list[str]] = {}
    _namespace_module_dict: dict[str, list[pkgutil.ModuleInfo]] = {}
    _namespace_sub_dict: dict[str, list[str]] = {}
    _module_dict: dict[str, list[str]] = {}
    _module_path_dict: dict[str, str] = {}
    _tree: dict[str, dict] = {}
    _stage_dict: dict[str, list[str]] = {}
    _base_stages: list[type] = []

    _skip_packages: list[str] = [
        "rail.projects",
        "rail.plotting",
        "rail.cli.rail_plot",
        "rail.cli.rail_project",
    ]

    _base_packages: list[str] = [
        "rail.core",
        "rail.stages",
        "rail.interfaces",
    ]

    _base_stages_names = [
        "CatClassifier",
        "PZClassifier",
        "CatEstimator",
        "CatInformer",
        "PzInformer",
        "CatSummarizer",
        "PZSummarizer",
        "SZPZSummarizer",
        "Degrader",
        "Noisifier",
        "Selector",
        "Modeler",
        "Creator",
        "PosteriorCalculator",
        "Evaluator",
    ]

    _module_api_options: str = """   :members:
   :undoc-members:
   :show-inheritance:
"""

    _module_no_index_api_options: str = """   :members:
   :undoc-members:
   :show-inheritance:
   :noindex:
"""

    _rail_core_api_options: str = """   :members:
   :undoc-members:
   :show-inheritance:
"""

    _package_api_options: str = """   :members:
   :undoc-members:
   :show-inheritance:
   :noindex:
"""

    @classmethod
    def list_rail_packages(cls) -> dict[str, pkgutil.ModuleInfo]:
        """List all the packages that are available in the RAIL ecosystem

        Returns
        -------
        dict[str,str]:
            Dict mapping the package names to the path to the package

        """
        cls._packages = {
            pkg.name: pkg
            for pkg in pkgutil.iter_modules(rail.__path__, rail.__name__ + ".")
        }
        return cls._packages

    @classmethod
    def print_rail_packages(cls) -> None:
        """Print all the packages that are available in the RAIL ecosystem"""
        if not cls._packages:  # pragma: no cover
            cls.list_rail_packages()
        for pkg_name, pkg in cls._packages.items():
            assert isinstance(pkg[0], importlib.machinery.FileFinder)
            path = pkg[0].path
            print(f"{pkg_name} @ {path}")

    @classmethod
    def list_rail_namespaces(cls) -> dict[str, list[str]]:
        """List all the namespaces within rail

        Returns
        -------
        dict[str, list[str]]:
            Dict mapping the namespaces to the paths contributing to
            each namespace
        """
        cls._namespace_path_dict.clear()

        for path_ in rail.__path__:
            namespaces = setuptools.find_namespace_packages(path_, exclude=["_*"])
            for namespace_ in namespaces:
                # exclude stuff that starts with 'examples_data'
                if namespace_.find("examples_data") == 0:
                    continue
                full_ns = f"rail.{namespace_}"
                if full_ns in cls._namespace_path_dict:  # pragma: no cover
                    cls._namespace_path_dict[full_ns].append(path_)
                else:
                    cls._namespace_path_dict[full_ns] = [path_]

                full_path = os.path.join(path_, namespace_.replace(".", "/"))
                cls._namespace_sub_dict[full_ns] = [
                    f"{full_ns}.{ns_}"
                    for ns_ in setuptools.find_namespace_packages(
                        full_path, exclude=["_*"]
                    )
                ]

        return cls._namespace_path_dict

    @classmethod
    def print_rail_namespaces(cls) -> None:
        """Print all the namespaces that are available in the RAIL ecosystem"""
        if not cls._namespace_path_dict:
            cls.list_rail_namespaces()
        for key, val in cls._namespace_path_dict.items():
            print(f"Namespace {key}")
            for vv in val:
                print(f"     {vv}")
        for key, val in cls._namespace_sub_dict.items():
            print(f"Namespace members {key}")
            for vv in val:
                print(f"     {vv}")

    @classmethod
    def list_rail_modules(cls) -> dict[str, str]:
        """List all modules within rail

        Returns
        -------
        dict[str, str]
            Dict mapping module names to their import paths
        """
        cls._module_dict.clear()
        cls._module_path_dict.clear()
        cls._namespace_module_dict.clear()
        if not cls._namespace_path_dict:  # pragma: no cover
            cls.list_rail_namespaces()
        for key, val in cls._namespace_path_dict.items():
            cls._namespace_module_dict[key] = []
            for vv in val:
                fullpath = os.path.join(vv, key.replace(".", "/")[5:])
                modules = list(pkgutil.iter_modules([fullpath], key + "."))

                for module_ in modules:
                    # Skip hidden files
                    if module_.name.find("._") >= 0:
                        continue
                    if module_.name in cls._module_dict:  # pragma: no cover
                        cls._module_dict[module_.name].append(key)
                    else:
                        cls._module_dict[module_.name] = [key]
                    cls._namespace_module_dict[key].append(module_)
                    assert isinstance(module_[0], importlib.machinery.FileFinder)
                    cls._module_path_dict[module_.name] = module_[0].path

        return cls._module_path_dict

    @classmethod
    def print_rail_modules(cls) -> None:
        """Print all the moduels that are available in the RAIL ecosystem"""
        if not cls._module_dict:
            cls.list_rail_modules()

        for key, val in cls._module_dict.items():
            print(f"Module {key}")
            for vv in val:
                print(f"     {vv}")

    @classmethod
    def build_rail_namespace_tree(cls) -> dict[str, dict]:
        """Build a tree of the namespaces and packages in rail

        Returns
        -------
        dict[str, list[dict]]:
           Tree of the namespaces and packages in rail
        """
        cls._tree.clear()
        if not cls._namespace_module_dict:  # pragma: no cover
            cls.list_rail_modules()

        if not cls._packages:  # pragma: no cover
            cls.list_rail_packages()

        # This is tricky, we are reconstrucing the source code tree
        # from the various import names by parsing the names and
        # identifying parents
        level_dict: dict[int, set[str]] = {}
        for key in cls._namespace_module_dict:
            count = key.count(".") - 1
            if count in level_dict:
                level_dict[count].add(key)
            else:
                level_dict[count] = {key}

        for key in cls._namespace_sub_dict:
            count = key.count(".") - 1
            level_dict[count].add(key)

        def _recursive_get(a_dict: dict, keys: list[str]) -> dict:
            ret_dict = a_dict
            use_keys = []
            for key_ in keys:
                use_keys.append(key_)
                full_key = ".".join(use_keys)
                full_key = f"rail.{full_key}"
                ret_dict = ret_dict[full_key]
            return ret_dict

        depth = max(level_dict.keys())

        for current_depth in range(depth + 1):
            for key in level_dict[current_depth]:
                subs = set()

                for module_ in cls._namespace_module_dict.get(key, []):
                    subs.add(module_.name)
                for ns_ in cls._namespace_sub_dict.get(key, []):
                    subs.add(ns_)

                a_dict: dict[str, dict] = {sub_: {} for sub_ in subs}

                parent_keys = key.split(".")[1 : current_depth + 1]
                parent_dict = _recursive_get(cls._tree, parent_keys)

                if key in parent_dict:
                    for kk, vv in a_dict.items():
                        parent_dict[key][kk] = vv
                else:
                    parent_dict[key] = a_dict

        return cls._tree

    @classmethod
    def pretty_print_tree(cls, the_dict: dict | None = None, indent: str = "") -> None:
        """Utility function to help print the namespace tree

        This can be called recurisvely to walk the tree structure, which has nested dicts

        Parameters
        ----------
        the_dict:
            Current dictionary to print, if None it will print cls._tree

        indent:
            Indentation string prepended to each line
        """
        if the_dict is None:  # pragma: no cover
            the_dict = cls._tree
        for key, val in the_dict.items():
            nsname = f"{key}"
            if nsname in cls._packages:
                pkg_type = "Package"
            elif nsname in cls._module_dict:
                pkg_type = "Module"
            else:
                pkg_type = "Namespace"

            print(f"{indent}{pkg_type} {nsname}")

            cls.pretty_print_tree(val, indent=indent + "    ")

    @classmethod
    def print_rail_namespace_tree(cls) -> None:
        """Print the namespace tree in a nice way"""
        if not cls._tree:
            cls.build_rail_namespace_tree()
        cls.pretty_print_tree(cls._tree)

    @classmethod
    def do_module_api_str(cls, basedir: str, key: str, options: str) -> None:
        """Build the api rst file for a rail module

        Parameters
        ----------
        basedir
            Directory to write file to

        key
            Name of the rail module

        options
            Pre-formatted autodoc options
        """

        api_pkg_toc = f"{key} module\n"
        api_pkg_toc += "-" * len(api_pkg_toc)

        api_pkg_toc += f"""

.. automodule:: {key}
{options}

"""
        print(f"Writing {key}.rst")
        with open(
            os.path.join(basedir, "api", f"{key}.rst"), "w", encoding="utf-8"
        ) as apitocfile:
            apitocfile.write(api_pkg_toc)

    @classmethod
    def do_pkg_api_rst(
        cls, basedir: str, key: str, val: dict, options: str, module_options: str
    ) -> None:
        """Build the api rst file for a rail package

        Parameters
        ----------
        basedir
            Directory to write file to

        key
            Name of the rail package

        val
            Namespace tree for the package

        options
            Pre-formatted autodoc options for the package

        module_options
            Pre-formatted autodoc options for modules
        """

        api_pkg_toc = f"{key} package\n"
        api_pkg_toc += "=" * len(api_pkg_toc)

        api_pkg_toc += f"""

.. automodule:: {key}
{options}

.. toctree::
   :maxdepth: 4
   :caption: {key}

"""

        for k2 in val:
            api_pkg_toc += f"   {k2}\n"
            cls.do_module_api_str(basedir, k2, module_options)

        print(f"Writing {key}.rst")
        with open(
            os.path.join(basedir, "api", f"{key}.rst"), "w", encoding="utf-8"
        ) as apitocfile:
            apitocfile.write(api_pkg_toc)

    @classmethod
    def do_namespace_api_rst(cls, basedir: str, key: str, val: dict) -> None:
        """Build the api rst file for a rail namespace

        Parameters
        ----------
        basedir
            Directory to write file to

        key
            Name of the rail namespace

        val:
            Namespace tree for the namespace
        """

        api_pkg_toc = f"{key} namespace\n"
        api_pkg_toc += "=" * len(api_pkg_toc)

        api_pkg_toc += """

.. py:module:: {key}

.. toctree::
   :maxdepth: 4
   :caption: {key}

{sub_packages}
"""

        sub_packages = ""
        for k2, v2 in val.items():
            if k2 in cls._skip_packages:  # pragma: no cover
                continue

            sub_packages += f"   {k2}\n"
            if k2 in cls._module_dict:
                cls.do_module_api_str(basedir, k2, cls._module_api_options)
                continue
            cls.do_namespace_api_rst(basedir, k2, v2)

        api_pkg_toc = api_pkg_toc.format(
            key=key,
            sub_packages=sub_packages,
        )

        print(f"Writing {key}.rst")
        with open(
            os.path.join(basedir, "api", f"{key}.rst"), "w", encoding="utf-8"
        ) as apitocfile:
            apitocfile.write(api_pkg_toc)

    @classmethod
    def do_api_rst(cls, basedir: str = ".") -> None:
        """Build the top-level API documentation

        Parameters
        ----------
        basedir
            Directory to write file to
        """

        if not cls._tree:  # pragma: no cover
            cls.build_rail_namespace_tree()

        apitoc = """

*************
Base Packages
*************

.. toctree::
   :maxdepth: 4
   :caption: Base Packages

{base_packages}


**********
Namespaces
**********

.. toctree::
   :maxdepth: 4
   :caption: Namespaces

{namespaces}


******************
Algorithm Packages
******************

.. toctree::
   :maxdepth: 4
   :caption: Algorithm Packages

{algorithm_packages}

"""
        try:
            os.makedirs(basedir)
        except Exception:
            pass

        try:
            os.makedirs(os.path.join(basedir, "api"))
        except Exception:  # pragma: no cover
            pass

        base_packages = ""
        namespaces = ""
        algorithm_packages = ""

        for key, val in cls._tree.items():
            nsname = f"{key}"
            nsfile = os.path.join("api", f"{nsname}")

            if nsname in cls._packages:
                # Skip rail_projects
                if nsname in cls._skip_packages:  # pragma: no cover
                    continue
                if nsname in cls._base_packages:
                    base_packages += f"    {nsfile}\n"
                    cls.do_pkg_api_rst(
                        basedir,
                        key,
                        val,
                        cls._rail_core_api_options,
                        cls._module_no_index_api_options,
                    )
                else:  # pragma: no cover
                    algorithm_packages += f"   {nsfile}\n"
                    cls.do_pkg_api_rst(
                        basedir,
                        key,
                        val,
                        cls._package_api_options,
                        cls._module_api_options,
                    )
            else:
                cls.do_namespace_api_rst(basedir, key, val)
                namespaces += f"   {nsfile}\n"

        apitoc = apitoc.format(
            base_packages=base_packages,
            namespaces=namespaces,
            algorithm_packages=algorithm_packages,
        )
        with open(
            os.path.join(basedir, "api.rst"), "w", encoding="utf-8"
        ) as apitocfile:
            apitocfile.write(apitoc)

    @classmethod
    def import_all_packages(cls, silent: bool = False) -> None:
        """Import all the packages that are available in the RAIL ecosystem"""
        pkgs = cls.list_rail_packages()
        for pkg in pkgs:
            try:
                _imported_module = importlib.import_module(pkg)
                if not silent:
                    print(f"Imported {pkg}")
            except Exception as msg:  # pragma: no cover
                if not silent:
                    print(f"Failed to import {pkg} because: {str(msg)}")

    @classmethod
    def attach_stages(cls, to_module: ModuleType) -> None:
        """Attach all the available stages to this module

        Parameters
        ----------
        to_module
            python module we are attaching stages to

        Notes
        -----
        This allow you to do:

        .. highlight:: python
        .. code-block:: python

            from rail.stages import *

        """
        from rail.core.stage import \
            RailStage  # pylint: disable=import-outside-toplevel

        cls._stage_dict.clear()
        cls._stage_dict["RailStage"] = []
        cls._base_stages.clear()

        n_base_classes = 0
        n_stages = 0

        for stage_name, stage_info in RailStage.incomplete_pipeline_stages.items():
            if stage_info[0] in [RailStage]:
                cls._stage_dict[stage_info[0].__name__] = []
            if stage_info[0].__name__ in cls._base_stages_names:
                cls._base_stages.append(stage_info[0])
                cls._stage_dict[stage_info[0].__name__] = []
                n_base_classes += 1

        for stage_name, stage_info in RailStage.pipeline_stages.items():
            if stage_info[0].__name__ in cls._base_stages_names:
                cls._base_stages.append(stage_info[0])
                cls._stage_dict[stage_info[0].__name__] = []
                n_base_classes += 1
                continue

            setattr(to_module, stage_name, stage_info[0])
            n_stages += 1

        for stage_name, stage_info in RailStage.pipeline_stages.items():
            baseclass = "RailStage"
            for possible_base in cls._base_stages:
                if issubclass(stage_info[0], possible_base):
                    baseclass = possible_base.__name__
                    break
            cls._stage_dict[baseclass].append(stage_name)

        print(
            f"Attached {n_base_classes} base classes and {n_stages} fully formed stages to rail.stages"
        )

    @classmethod
    def print_rail_stage_dict(cls) -> None:
        """Print an dict of all the RailSages organized by their base class"""
        from rail.core.stage import \
            RailStage  # pylint: disable=import-outside-toplevel

        for key, val in cls._stage_dict.items():
            print(f"BaseClass {key}")
            for vv in val:
                stage_class = RailStage.pipeline_stages[vv][0]
                print(f"  {vv} {stage_class.__module__}.{stage_class.__name__}")

    @classmethod
    def do_stage_type_api_rst(cls, basedir: str = ".") -> None:
        """Genarate the rst files for the stage tpye documentation"""
        from rail.core.stage import \
            RailStage  # pylint: disable=import-outside-toplevel

        os.makedirs(os.path.join(basedir, "api"), exist_ok=True)

        for key, val in cls._stage_dict.items():
            if key in RailStage.incomplete_pipeline_stages:
                base_class = RailStage.incomplete_pipeline_stages[key][0]
            else:
                base_class = RailStage.pipeline_stages[key][0]

            api_stage_type = f"{key} Stage Type\n"
            api_stage_type += "*" * len(api_stage_type)
            api_stage_type += "\n"

            api_stage_type += (
                f".. autoclass:: {base_class.__module__}.{base_class.__name__}\n"
            )
            api_stage_type += "    :noindex:\n\n"

            subheader = f"{key} Sub-Classes\n"
            api_stage_type += subheader
            api_stage_type += "=" * len(subheader)
            api_stage_type += "\n"

            for vv in val:
                stage_class = RailStage.pipeline_stages[vv][0]

                api_stage_type += (
                    f".. autoclass:: {stage_class.__module__}.{stage_class.__name__}\n"
                )
                api_stage_type += "   :noindex:\n\n"

            print(f"Writing {key}_stage_type.rst")
            with open(
                os.path.join(basedir, "api", f"{key}_stage_type.rst"),
                "w",
                encoding="utf-8",
            ) as api_stage_type_file:
                api_stage_type_file.write(api_stage_type)

        header = "Types of RAIL stages\n"
        api_stage_type_index = ""
        api_stage_type_index += "*" * len(header)
        api_stage_type_index += "\n"
        api_stage_type_index += header
        api_stage_type_index += "*" * len(header)
        api_stage_type_index += "\n\n"

        api_stage_type_index += ".. toctree::\n"
        api_stage_type_index += "    :maxdepth: 4\n\n"

        for key in cls._stage_dict:
            api_stage_type_index += f"    {key}_stage_type\n"

        print("Writing stage_types.rst")
        with open(
            os.path.join(basedir, "api", "stage_types.rst"), "w", encoding="utf-8"
        ) as api_stage_type_index_file:
            api_stage_type_index_file.write(api_stage_type_index)



import os

import numpy as np
import pytest
import scipy.special

from rail.core.data import PqHandle, TableHandle, QPHandle
from rail.core.stage import RailStage
from rail.estimation.algos import random_gauss, train_z, gaussian_pz
from rail.utils.path_utils import RAILDIR
from rail.utils.testing_utils import one_algo

sci_ver_str = scipy.__version__.split(".")


DS = RailStage.data_store
DS.__class__.allow_overwrite = True


def test_random_pz() -> None:
    train_config_dict: dict = {}
    estim_config_dict = {
        "rand_width": 0.025,
        "rand_zmin": 0.0,
        "rand_zmax": 3.0,
        "calc_summary_stats": True,
        "nzbins": 301,
        "hdf5_groupname": "photometry",
        "model": "None",
        "seed": 42,
    }
    zb_expected = np.array(
        [2.322, 1.317, 2.576, 2.092, 0.283, 2.927, 2.283, 2.358, 0.384, 1.351]
    )
    train_algo = random_gauss.RandomGaussInformer
    pz_algo = random_gauss.RandomGaussEstimator
    results, _, _ = one_algo(
        "RandomPZ", train_algo, pz_algo, train_config_dict, estim_config_dict
    )
    assert np.isclose(results.ancil["zmode"], zb_expected).all()
    try:
        os.remove("model.pkl")
    except FileNotFoundError:  # pragma: no cover
        pass


def test_train_pz() -> None:
    train_config_dict = dict(
        zmin=0.0,
        zmax=3.0,
        nzbins=301,
        hdf5_groupname="photometry",
        model="model_train_z.tmp",
    )
    estim_config_dict = dict(hdf5_groupname="photometry", model="model_train_z.tmp")

    zb_expected = np.repeat(0.1445183, 10)
    pdf_expected = np.zeros(shape=(301,))
    pdf_expected[10:16] = [7, 23, 8, 23, 26, 13]
    train_algo = train_z.TrainZInformer
    pz_algo = train_z.TrainZEstimator
    results, rerun_results, _ = one_algo(
        "TrainZ", train_algo, pz_algo, train_config_dict, estim_config_dict
    )

    assert (results.ancil['distribution_type']==0).all()
    assert np.isclose(results.ancil["zmode"], zb_expected).all()
    assert np.isclose(results.ancil["zmode"], rerun_results.ancil["zmode"]).all()

    try:
        os.remove("model.pkl")
    except FileNotFoundError:  # pragma: no cover
        pass


def test_train_pz_with_wrong_columns_path() -> None:
    DS.clear()
    DS.__class__.allow_overwrite = False

    datapath_pq = os.path.join(
        RAILDIR, "rail", "examples_data", "testdata", "test_dc2_training_9816.pq"
    )

    # ! create training data to be a data handle with path only
    # ! however it seems that with set_data() one always reads in the data
    # ! hence the way we make the data here:
    training_data1 = DS.add_handle("pq1", PqHandle, path=datapath_pq)

    train_config_dict = dict(
        zmin=0.0,
        zmax=3.0,
        nzbins=301,
        hdf5_groupname=None,
        model="model_train_z.tmp",
        redshift_col="REDSHIFT",
    )

    train_algo = train_z.TrainZInformer
    train_pz = train_algo.make_stage(**train_config_dict)
    train_pz._get_stage_columns()
    with pytest.raises(KeyError):
        # testing the case where only path exist
        train_pz._check_column_names(training_data1, train_pz.stage_columns)


def test_train_pz_with_wrong_columns_nogroupname() -> None:
    DS.clear()
    DS.__class__.allow_overwrite = False

    datapath_pq = os.path.join(
        RAILDIR, "rail", "examples_data", "testdata", "test_dc2_training_9816.pq"
    )

    training_data2 = DS.read_file("pq2", PqHandle, path=datapath_pq)

    train_config_dict = dict(
        zmin=0.0,
        zmax=3.0,
        nzbins=301,
        hdf5_groupname=None,
        model="model_train_z.tmp",
        redshift_col="REDSHIFT",
    )

    train_algo = train_z.TrainZInformer
    train_pz = train_algo.make_stage(**train_config_dict)
    train_pz._get_stage_columns()
    with pytest.raises(KeyError):
        # testing the case where data handle has no hdf5 group name
        train_pz._check_column_names(training_data2, train_pz.stage_columns)


def test_train_pz_with_wrong_columns_table() -> None:
    DS.clear()
    DS.__class__.allow_overwrite = False

    datapath_pq = os.path.join(
        RAILDIR, "rail", "examples_data", "testdata", "test_dc2_training_9816.pq"
    )

    # ! create training data to be a table
    # ! however it seems that with set_data() one always reads in the data handle
    # ! hence the way we make the data here:
    training_data2 = DS.read_file("pq2", PqHandle, path=datapath_pq)
    training_data3 = training_data2.data

    train_config_dict = dict(
        zmin=0.0,
        zmax=3.0,
        nzbins=301,
        hdf5_groupname=None,
        model="model_train_z.tmp",
        redshift_col="REDSHIFT",
    )

    train_algo = train_z.TrainZInformer
    train_pz = train_algo.make_stage(**train_config_dict)
    train_pz._get_stage_columns()
    with pytest.raises(KeyError):
        # testing the case where data is a table
        train_pz._check_column_names(training_data3, train_pz.stage_columns)


def test_train_pz_with_wrong_columns_table_wgroupname() -> None:
    DS.clear()
    DS.__class__.allow_overwrite = False

    traindata = os.path.join(
        RAILDIR, "rail/examples_data/testdata/training_100gal.hdf5"
    )

    # ! create training data to be a table
    # ! however it seems that with set_data() one always reads in the data handle
    # ! hence the way we make the data here:
    training_data2 = DS.read_file("training_data", TableHandle, traindata)
    training_data3 = training_data2.data

    train_config_dict = dict(
        zmin=0.0,
        zmax=3.0,
        nzbins=301,
        hdf5_groupname="photometry",
        model="model_train_z.tmp",
        redshift_col="REDSHIFT",
    )

    train_algo = train_z.TrainZInformer
    train_pz = train_algo.make_stage(**train_config_dict)
    train_pz._get_stage_columns()
    with pytest.raises(KeyError):
        # testing the case where data is a table
        train_pz._check_column_names(training_data3, train_pz.stage_columns)



def test_gaussian_pz() -> None:
        
    DS.clear()
    DS.__class__.allow_overwrite = False

    input_path = os.path.join(
        RAILDIR, "rail/examples_data/testdata/output_BPZ_lite.hdf5"
    )

    input_data = DS.read_file("input_data", QPHandle, input_path)

    inform_gauss_pz = gaussian_pz.GaussianPzInformer.make_stage()
    model = inform_gauss_pz.inform(input_data)
    
    estimate_gauss_pz = gaussian_pz.GaussianPzEstimator.make_stage(model=model)
    outdata = estimate_gauss_pz.estimate(input_data)
    assert outdata.data.npdf == 10
    



"""Degrader that applies a cut to given columns."""

from numbers import Number
from typing import Any

import numpy as np
from ceci.config import StageParameter as Param

from rail.creation.selector import Selector


class QuantityCut(Selector):
    """Degrader that applies a cut to the given columns.

    Note that if a galaxy fails any of the cuts on any one of its columns, that
    galaxy is removed from the sample.
    """

    name = "QuantityCut"
    config_options = Selector.config_options.copy()
    config_options.update(
        cuts=Param(dict, required=True, msg="Cuts to apply"),
    )

    def __init__(self, args: Any, **kwargs: Any) -> None:
        """Constructor.

        Performs standard Degrader initialization as well as defining the cuts
        to be applied.
        """
        super().__init__(args, **kwargs)
        self.cuts: dict | None = None
        self.set_cuts(self.config["cuts"])

    def set_cuts(self, cuts: dict) -> None:
        """Defines the cuts to be applied.

        Parameters
        ----------
        cuts : dict
            A dictionary of cuts to make on the data

        Notes
        -----
        The cut keys should be the names of columns you wish to make cuts on.

        The cut values should be either:
        - a number, which is the maximum value. E.g. if the dictionary
        contains "i": 25, then values of i > 25 are cut from the sample.
        - an iterable, which is the range of acceptable values. E.g. if the
        dictionary contains "redshift": (1.5, 2.3), then redshifts outside that
        range are cut from the sample.
        """
        # check that cuts is a dictionary
        if not isinstance(cuts, dict):  # pragma: no cover
            raise TypeError("cuts must be a dictionary.")

        # validate all the cuts and standardize format in dictionary
        self.cuts = {}
        for quantity, cut in cuts.items():
            bad_cut_msg = (
                f"Cut for {quantity} must be a number or an iterable of (min, max)"
            )
            # if single number is provided, save that as the maximum
            if isinstance(cut, Number):
                self.cuts[quantity] = (-np.inf, cut)
            # else, if it's an iterable...
            elif hasattr(cut, "__iter__"):
                # if the iterable is a string or dict, raise error
                if isinstance(cut, (dict, str)):
                    raise TypeError(bad_cut_msg)
                # if the length of the iterable isn't 2, raise error
                if len(cut) != 2:
                    raise ValueError(bad_cut_msg)
                # check that both of the cut values are a number
                for c in cut:
                    if not isinstance(c, Number):
                        raise TypeError(bad_cut_msg)
                # check that the cuts are (min, max)
                if cut[0] > cut[1]:
                    raise ValueError(
                        f"The max for {quantity} must be greater than the min."
                    )
                # if all of these checks passed, save the cuts
                self.cuts[quantity] = (cut[0], cut[1])
            # if the cut isn't a number or iterable, it's the wrong type
            else:
                raise TypeError(bad_cut_msg)

    def _select(self) -> np.ndarray:
        """Applies cuts.

        Notes
        -----
        Gets the input data from the data store under this stage's 'input' tag.

        Puts the data into the data store under this stage's 'output' tag.
        """
        data = self.get_data("input")

        assert self.cuts is not None

        # get overlap of columns from data and columns on which to make cuts
        columns = set(self.cuts.keys()).intersection(data.columns)

        if len(columns) == 0:  # pragma: no cover
            return np.ones(len(data), dtype=int)
        # generate a pandas query from the cuts
        query = [
            f"{col} > {self.cuts[col][0]} & {col} < {self.cuts[col][1]}"
            for col in columns
        ]
        query_st = " & ".join(query)
        out_indices = data.query(query_st).index.values
        out_mask = np.zeros(len(data), dtype=int)
        out_mask[out_indices] = 1
        return out_mask

    def __repr__(self) -> str:  # pragma: no cover
        """Pretty print this object."""
        printMsg = "Degrader that applies the following cuts to a pandas DataFrame:\n"
        printMsg += "{column: (min, max), ...}\n"
        printMsg += self.cuts.__str__()
        return printMsg



#!/usr/bin/env python
# coding: utf-8

from rail.core.stage import RailPipeline, RailStage
# Various rail modules
from rail.estimation.algos.train_z import TrainZEstimator, TrainZInformer
from rail.evaluation.single_evaluator import SingleEvaluator


class TrainZPipeline(RailPipeline):
    default_input_dict: dict[str, str] = dict(
        input_train="dummy.in",
        input_test="dummy.in",
    )

    def __init__(self) -> None:
        RailPipeline.__init__(self)

        DS = RailStage.data_store
        DS.__class__.allow_overwrite = True

        self.inform_trainz = TrainZInformer.build(
            aliases=dict(input="input_train"),
            hdf5_groupname="",
        )

        self.estimate_trainz = TrainZEstimator.build(
            aliases=dict(input="input_test"),
            connections=dict(
                model=self.inform_trainz.io.model,
            ),
            hdf5_groupname="",
        )

        self.evalute_trainz = SingleEvaluator.build(
            aliases=dict(truth="input_test"),
            connections=dict(
                input=self.estimate_trainz.io.output,
            ),
            point_estimates=["mode"],
            truth_point_estimates=["redshift"],
            metrics=["all"],
            metric_config=dict(brier=dict(limits=[0.0, 3.5])),
            exclude_metrics=["rmse", "ks", "kld", "cvm", "ad", "rbpe", "outlier"],
            hdf5_groupname="",
        )



import numpy as np
from scipy import stats
from ceci.config import StageParameter as Param

from rail.core.common_params import SHARED_PARAMS
from rail.core.data import TableLike, DataHandle, Hdf5Handle, TableHandle
from rail.core.stage import RailStage


class KDEBinOverlap(RailStage):
    name = "KDEBinOverlap"
    inputs = [("truth", TableHandle), ("bin_index", Hdf5Handle)]
    outputs = [("output", Hdf5Handle)]

    config_options = RailStage.config_options.copy()
    config_options.update(
        hdf5_groupname=Param(
            str, "", required=False, msg="HDF5 Groupname for truth table."
        ),
        redshift_col=SHARED_PARAMS,
        bin_name=Param(
            str,
            "class_id",
            required=False,
            msg="Groupname for the tomographic bin index in the hdf5 handle",
        ),
    )
    # metric_base_class = Evaluator

    def evaluate(self, bin_index: TableLike, truth: TableLike) -> DataHandle:
        self.set_data("bin_index", bin_index)
        self.set_data("truth", truth)

        self.run()
        return self.get_handle("output")

    def run(self) -> None:
        if self.config.hdf5_groupname is None: # pragma: no cover
            true_redshifts = self.get_handle("truth").data[
                self.config.redshift_col
            ]  # 1D array of redshifts
        else: 
            true_redshifts = self.get_handle("truth").data[self.config.hdf5_groupname][
                self.config.redshift_col
            ]  # 1D array of redshifts
        bin_indices = self.get_handle("bin_index").data[
            self.config.bin_name
        ]  # 1D array of bin indices

        unique_bins = np.unique(bin_indices)
        unique_bins = unique_bins[unique_bins >= 0]  # to exclude the -99
        N = len(unique_bins)
        overlap_matrix = np.zeros((N, N))

        kde_dict = {}
        for bin_id in unique_bins:
            redshifts_in_bin = true_redshifts[bin_indices == bin_id]
            if len(redshifts_in_bin) > 1:
                kde = stats.gaussian_kde(redshifts_in_bin)
                kde_dict[bin_id] = kde

        for i, bin_i in enumerate(unique_bins):
            overlap_matrix[i][i] = 1.0
            for j, bin_j in enumerate(unique_bins):
                if j >= i:
                    continue

                kde_i = kde_dict.get(bin_i, None)
                kde_j = kde_dict.get(bin_j, None)

                if kde_i is not None and kde_j is not None:
                    # Define the evaluation grid based on the union of both bin samples
                    eval_grid = np.linspace(
                        min(true_redshifts), max(true_redshifts), 1000
                    )
                    p_i = kde_i(eval_grid)
                    p_j = kde_j(eval_grid)

                    # Compute overlap as the integral of the minimum of both distributions
                    overlap = np.trapz(np.minimum(p_i, p_j), eval_grid)
                    overlap_matrix[i, j] = overlap
                    overlap_matrix[j, i] = overlap  # Symmetric matrix

        self.add_handle("output", data=overlap_matrix)



# pylint: disable=no-member
import numpy as np
import pytest
import qp

from rail.estimation.estimator import CatEstimator


def test_custom_point_estimate() -> None:
    """This test checks to make sure that the inheritance mechanism is working
    for child classes of `CatEstimator`.
    """

    MEANING_OF_LIFE = 42.0

    class TestEstimator(CatEstimator):
        name = "TestEstimator"

        def _calculate_mode_point_estimate(
            self, qp_dist: None = None, grid: None = None
        ) -> np.ndarray:
            return np.ones(100) * MEANING_OF_LIFE

    config_dict = {"calculated_point_estimates": ["zmode"]}

    test_estimator = TestEstimator.make_stage(name="test", **config_dict)

    locs = 2 * (np.random.uniform(size=(100, 1)) - 0.5)
    scales = 1 + 0.2 * (np.random.uniform(size=(100, 1)) - 0.5)
    test_ensemble = qp.Ensemble(
        qp.stats.norm, data=dict(loc=locs, scale=scales)
    )  # pylint: disable=no-member

    result = test_estimator.calculate_point_estimates(test_ensemble)

    assert np.all(result.ancil["zmode"] == MEANING_OF_LIFE)


def test_basic_point_estimate() -> None:
    """This test checks to make sure that all the basic point estimates are
    executed when requested in the configuration dictionary.
    """

    config_dict = {
        "calculated_point_estimates": ["zmean", "zmedian", "zmode", "zbest"],
        "zmin": 0.0,
        "zmax": 3.0,
        "nzbins": 301,
    }

    test_estimator = CatEstimator.make_stage(name="test", **config_dict)

    locs = 2 * (np.random.uniform(size=(100, 1)) - 0.5)
    scales = 1 + 0.2 * (np.random.uniform(size=(100, 1)) - 0.5)
    test_ensemble = qp.Ensemble(
        qp.stats.norm, data=dict(loc=locs, scale=scales)
    )  # pylint: disable=no-member
    result = test_estimator.calculate_point_estimates(test_ensemble, None)

    # note: we're not interested in testing the values of point estimates,
    # just that they were added to the ancillary data.
    assert "zmode" in result.ancil
    assert "zmedian" in result.ancil
    assert "zmean" in result.ancil
    assert "zbest" in result.ancil


def test_mode_no_grid() -> None:
    """This exercises the KeyError logic in `_calculate_mode_point_estimate`."""
    config_dict = {"zmin": 0.0, "nzbins": 100, "calculated_point_estimates": ["mode"]}

    test_estimator = CatEstimator.make_stage(name="test", **config_dict)

    with pytest.raises(AttributeError) as excinfo:
        _ = test_estimator.calculate_point_estimates(None, None)

    assert "object has no attribute" in str(excinfo.value)


def test_mode_no_point_estimates() -> None:
    """This exercises the KeyError logic in `_calculate_mode_point_estimate`."""
    config_dict = {"zmin": 0.0, "nzbins": 100}

    test_estimator = CatEstimator.make_stage(name="test", **config_dict)

    locs = 2 * (np.random.uniform(size=(100, 1)) - 0.5)
    scales = 1 + 0.2 * (np.random.uniform(size=(100, 1)) - 0.5)
    test_ensemble = qp.Ensemble(
        qp.stats.norm, data=dict(loc=locs, scale=scales)
    )  # pylint: disable=no-member

    output_ensemble = test_estimator.calculate_point_estimates(test_ensemble, None)

    assert output_ensemble.ancil is None


def test_keep_existing_ancil_data() -> None:
    """Make sure that we don't overwrite the ancil data if it already exists."""
    config_dict = {
        "zmin": 0.0,
        "zmax": 3.0,
        "nzbins": 100,
        "calculated_point_estimates": ["zmode"],
    }

    test_estimator = CatEstimator.make_stage(name="test", **config_dict)

    locs = 2 * (np.random.uniform(size=(100, 1)) - 0.5)
    scales = 1 + 0.2 * (np.random.uniform(size=(100, 1)) - 0.5)
    test_ensemble = qp.Ensemble(qp.stats.norm, data=dict(loc=locs, scale=scales))

    test_ensemble.set_ancil({"foo": np.zeros(100)})

    output_ensemble = test_estimator.calculate_point_estimates(test_ensemble, None)

    assert "foo" in output_ensemble.ancil
    assert np.all(output_ensemble.ancil["foo"] == 0.0)
    assert len(output_ensemble.ancil["foo"]) == 100


def test_write_new_ancil_data() -> None:
    """Make sure that we don't overwrite the ancil data if it already exists."""
    config_dict = {
        "zmin": 0.0,
        "zmax": 3.0,
        "nzbins": 100,
        "calculated_point_estimates": ["zmode"],
    }

    test_estimator = CatEstimator.make_stage(name="test", **config_dict)

    locs = 2 * (np.random.uniform(size=(100, 1)) - 0.5)
    scales = 1 + 0.2 * (np.random.uniform(size=(100, 1)) - 0.5)
    test_ensemble = qp.Ensemble(qp.stats.norm, data=dict(loc=locs, scale=scales))

    test_ensemble.set_ancil({"foo": np.zeros(100)})

    output_ensemble = test_estimator.calculate_point_estimates(test_ensemble, None)

    assert "zmode" in output_ensemble.ancil
    assert len(output_ensemble.ancil["zmode"]) == 100



import os
from types import GeneratorType

import pytest

from rail.core.data import Hdf5Handle, ModelHandle, TableHandle
from rail.core.stage import RailStage
from rail.tools.table_tools import ColumnMapper, RowSelector, TableConverter
from rail.utils.path_utils import RAILDIR, find_rail_file


def test_find_rail_file() -> None:
    afile = find_rail_file(
        os.path.join("examples_data", "testdata", "test_dc2_training_9816.pq")
    )
    assert afile
    with pytest.raises(ValueError):
        _not_a_file = find_rail_file("not_a_file")


def test_util_stages() -> None:
    DS = RailStage.data_store
    DS.clear()
    datapath = os.path.join(
        RAILDIR, "rail", "examples_data", "testdata", "test_dc2_training_9816.pq"
    )

    data = DS.read_file("data", TableHandle, datapath)

    table_conv = TableConverter.make_stage(name="conv", output_format="numpyDict")
    col_map = ColumnMapper.make_stage(name="col_map", columns={})
    row_sel = RowSelector.make_stage(name="row_sel", start=1, stop=15)

    with pytest.raises(KeyError) as _errinfo:
        table_conv.get_handle("nope", allow_missing=False)

    _conv_data = table_conv(data)
    mapped_data = col_map(data)
    _sel_data = row_sel(mapped_data)

    row_sel_2 = RowSelector.make_stage(name="row_sel_2", start=1, stop=15)
    row_sel_2.set_data("input", mapped_data.data)
    handle = row_sel_2.get_handle("input")

    row_sel_3 = RowSelector.make_stage(
        name="row_sel_3", input=handle.path, start=1, stop=15
    )
    row_sel_3.set_data("input", None, do_read=True)

    for stage in [table_conv, col_map, row_sel]:
        os.remove(stage.get_output(stage.get_aliased_tag("output"), final_name=True))


def test_set_data_nonexistent_file() -> None:
    """Create an instance of a child class of RailStage. Exercise the `set_data`
    method and pass in a path to a nonexistent file. A `FileNotFound` exception
    should be raised.
    """

    col_map = ColumnMapper.make_stage(name="col_map", columns={})
    with pytest.raises(FileNotFoundError) as err:
        col_map.set_data("input", None, path="./bad_directory/no_file.py")
        assert "Unable to find file" in err.context


def test_set_data_real_file() -> None:
    """Create an instance of a child class of RailStage. Exercise the `set_data`
    method and pass in a path to model. The output of set_data should be `None`.
    """
    DS = RailStage.data_store
    DS.clear()
    model_path = os.path.join(
        RAILDIR,
        "rail",
        "examples_data",
        "estimation_data",
        "data",
        "CWW_HDFN_prior.pkl",
    )
    DS.add_data("model", None, ModelHandle, path=model_path)

    col_map = ColumnMapper.make_stage(name="col_map", columns={})

    ret_val = col_map.set_data("model", None, path=model_path, do_read=False)

    assert ret_val is None


def test_data_hdf5_iter() -> None:
    DS = RailStage.data_store
    DS.clear()

    datapath = os.path.join(
        RAILDIR, "rail", "examples_data", "testdata", "test_dc2_training_9816.hdf5"
    )

    # data = DS.read_file('data', TableHandle, datapath)
    th = Hdf5Handle("data", path=datapath)
    x = th.iterator(groupname="photometry", chunk_size=1000)

    assert isinstance(x, GeneratorType)
    for i, xx in enumerate(x):
        assert xx[0] == i * 1000
        assert xx[1] - xx[0] <= 1000

    _data = DS.read_file("input", TableHandle, datapath)
    cm = ColumnMapper.make_stage(
        input=datapath,
        chunk_size=1000,
        hdf5_groupname="photometry",
        columns=dict(id="bob"),
    )
    x = cm.input_iterator("input")

    assert isinstance(x, GeneratorType)

    for i, xx in enumerate(x):
        assert xx[0] == i * 1000
        assert xx[1] - xx[0] <= 1000
